#!/usr/bin/env python3

"""Send query to multiple FHIR endpoints, consume all pages, write results to stdout."""

import json
import logging
import os
import urllib.parse as urlparse

import click
from click_loglevel import LogLevel
from fhirclient.server import FHIRJSONMimeType

from anvil.clients.fhir_client import DispatchingFHIRClient
from anvil.clients.smart_auth import GoogleFHIRAuth
from anvil.etl.anvil_etl import LOG_FORMAT

logging.basicConfig(
    format='%(asctime)s %(levelname)-8s %(threadName)s %(message)s',
    level=logging.INFO,
    datefmt='%Y-%m-%d %H:%M:%S')

logger = logging.getLogger(__name__)


def worker(server, path=None, no_sign=False, headers={}):
    """Callback handler"""

    def fetch(url, _headers):
        """Low level GET, follow next links.
         see https://github.com/smart-on-fhir/client-py/blob/master/fhirclient/server.py#L174."""
        header_defaults = {
            'Accept': FHIRJSONMimeType,
            'Accept-Charset': 'UTF-8',
        }
        # merge in user headers with defaults
        header_defaults.update(_headers)
        _headers = header_defaults
        if not no_sign and server.auth is not None and server.auth.can_sign_headers():
            _headers = server.auth.signed_headers(_headers)
        # perform the request but intercept 401 responses, raising our own Exception
        res = server.session.get(url, headers=_headers)
        server.raise_for_status(res)
        __json = res.json()
        __next = None
        if 'link' in __json:
            _links = {lnk['relation']: lnk['url'] for lnk in __json['link']}
            if 'next' in _links:
                __next = _links['next']
        return __json, __next

    # main
    assert path, "caller MUST pass path"
    if path.startswith('/'):
        path = path[1:]
    initial_url = urlparse.urljoin(server.base_uri, path)
    _url = initial_url
    while _url:
        (_json, _url) = fetch(_url, _headers=headers)
        print(json.dumps(_json, separators=(',', ':')), flush=True)


def _dispatch(project, location, dataset, all_data_stores, requested_data_store, path, token):
    """Manufacture URLs, dispatch to thread."""

    all_data_stores = all_data_stores.split(',')
    if requested_data_store.lower() == 'dispatch':
        data_stores = all_data_stores
    else:
        # assert requested_data_store in all_data_stores, f"{requested_data_store} not in {all_data_stores}"
        data_stores = [requested_data_store]

    api_bases = [f"https://healthcare.googleapis.com/v1beta1/projects/{project}/locations/{location}/datasets/{dataset}/fhirStores/{ds.strip()}/fhir" for ds in data_stores if len(ds.strip()) > 0]

    settings = {
        'app_id': 'my_web_app',
        'api_bases': api_bases
    }
    client = DispatchingFHIRClient(settings=settings, auth=GoogleFHIRAuth(access_token=token))
    client.prepare()
    assert client.ready, "server should be ready"
    assert project in client.server.base_uri
    client.dispatch(worker, path=path)


def _get_data_stores():
    return os.environ.get('GOOGLE_DATASTORES', None)


def _get_data_stores_help():
    data_stores = _get_data_stores()
    if data_stores:
        return ""
    return """source /dev/stdin <<< `anvil_etl utility env`  """


def _get_token():
    return os.environ.get('TOKEN', None)


def _get_token_help():
    token = _get_token()
    if token:
        return f"{token[0:10]}..."
    return """export TOKEN=$(gcloud auth application-default print-access-token)"""


class DefaultCommand(click.Command):
    def parse_args(self, ctx, args):
        if len(args) == 1:
            args.insert(0, 'dispatch')
        super(DefaultCommand, self).parse_args(ctx, args)


@click.command(cls=DefaultCommand)
@click.option('--project', default=os.environ.get('FHIR_PROJECT', None), show_default=True, help='env var. FHIR_PROJECT')
@click.option('--location', default=os.environ.get('GOOGLE_LOCATION', None), show_default=True, help='env var. GOOGLE_LOCATION')
@click.option('--dataset', default=os.environ.get('GOOGLE_DATASET', None), show_default=True, help='env var. GOOGLE_DATASET')
@click.option('--token', default=_get_token(), help=f'env var. TOKEN [default: {_get_token_help()}]')
@click.option('--data_stores', default=_get_data_stores(), show_default=True, help=f'env var. GOOGLE_DATASTORES. [default: {_get_data_stores_help()}]')
@click.option("-l", "--log-level", type=LogLevel(), default=logging.INFO)
@click.argument('data_store', required=False, default='dispatch')
@click.argument('path', required=False)
def cli(project, location, dataset, token, data_stores, data_store, path, log_level):
    """Utility to manufacture FHIR URLs, dispatch path via threads.

        data_store: FHIR data store name.  'dispatch' will send to all data stores.
        path: FHIR compliant path i.e. http{s}://server{/path} see https://www.hl7.org/fhir/http.html#general
    """
    logging.basicConfig(level=log_level, format=LOG_FORMAT, force=True)
    logger.setLevel(log_level)
    if not data_stores:
        logger.error("Please set data_stores parameter")
    logger.debug(f"data_store:{data_store} path:{path}")
    assert logging.getLogger(__name__).level != logging.NOTSET, f"Should be {log_level}"
    _dispatch(project, location, dataset, data_stores, data_store, path, token)


if __name__ == '__main__':
    cli()



import os
import pkg_resources
import yaml
import click
from collections import OrderedDict

DEFAULT_OUTPUT_PATH = os.environ.get('OUTPUT_PATH', './DATA')

DEFAULT_GEN3_CREDENTIALS_PATH = os.path.expanduser('~/.gen3/credentials.json')

DEFAULT_FHIR_PROJECT = os.environ.get('FHIR_PROJECT', None)


def read_config(path=None):
    """Read config, if no path, read from installed resource."""

    if not path:
        resource_package = 'anvil.etl'
        path = 'utilities/config.yaml'  # Do not use os.path.join()
        config_file = pkg_resources.resource_stream(resource_package, path)
    else:
        config_file = open(path)

    config = yaml.load(config_file, Loader=yaml.SafeLoader)
    assert sorted(config.keys()) == ['consortiums', 'mapping'], f"Config missing expected keys. {config.keys()} "

    return config


class NaturalOrderGroup(click.Group):

    def __init__(self, name=None, commands=None, **attrs):
        super(NaturalOrderGroup, self).__init__(
            name=name, commands=None, **attrs)
        if commands is None:
            commands = OrderedDict()
        elif not isinstance(commands, OrderedDict):
            commands = OrderedDict(commands)
        self.commands = commands

    def list_commands(self, ctx):
        return self.commands.keys()


import subprocess
from collections import namedtuple
import os
import logging
logger = logging.getLogger(__name__)

PROPERTIES_LIST = """
    FHIR_PROJECT_NAME 
    GOOGLE_LOCATION 
    FHIR_PROJECT 
    GOOGLE_DATASET 
    TOKEN 
    GOOGLE_DATASTORES 
    GOOGLE_DATASTORE 
    GOOGLE_BUCKET 
    OUTPUT_PATH 
    IMPLEMENTATION_GUIDE_PATH""".strip().split()


def run_cmd(command_line):
    """Run a command line, return stdout."""
    try:
        logger.debug(command_line)
        return subprocess.check_output(command_line, shell=True).decode("utf-8").rstrip()
    except Exception as e:
        logger.error(e)
        raise e


def ensure_env_variables():
    """Fill in environment variables, ensure FHIR context."""

    variables = {property_: os.environ.get(property_) for property_ in PROPERTIES_LIST if os.environ.get(property_, None)}

    variables["FHIR_PROJECT_NAME"] = variables.get("FHIR_PROJECT_NAME", 'fhir-test-16')
    variables["FHIR_PROJECT"] = variables.get("FHIR_PROJECT", 'fhir-test-16-342800')
    variables["GOOGLE_LOCATION"] = variables.get("GOOGLE_LOCATION", 'us-west2')
    variables["GOOGLE_DATASET"] = variables.get("GOOGLE_DATASET", 'anvil-test')
    variables["GOOGLE_DATASTORE"] = variables.get("GOOGLE_DATASTORE", 'public')
    variables["OUTPUT_PATH"] = variables.get("OUTPUT_PATH", './DATA')
    variables["IMPLEMENTATION_GUIDE_PATH"] = variables.get("IMPLEMENTATION_GUIDE_PATH", f'{variables["OUTPUT_PATH"]}/fhir/IG')

    if not variables.get('FHIR_PROJECT', None):
        variables['FHIR_PROJECT'] = run_cmd(f'gcloud projects list --filter=name={variables["FHIR_PROJECT_NAME"]} --format="value(projectId)"')

    if not variables.get("GOOGLE_BUCKET", None):
        variables["GOOGLE_BUCKET"] = variables['FHIR_PROJECT']

    for k in ["GOOGLE_DATASET", "GOOGLE_LOCATION", 'FHIR_PROJECT', "GOOGLE_BUCKET"]:
        assert variables[k], f"Should have set variable {k}"

    if not variables.get('GOOGLE_DATASTORES', None):
        variables['GOOGLE_DATASTORES'] = run_cmd(f'gcloud beta healthcare fhir-stores list --dataset={variables["GOOGLE_DATASET"]}  --location={variables["GOOGLE_LOCATION"]}  --format="table[no-heading](ID)"')
    if isinstance(variables['GOOGLE_DATASTORES'], str):
        variables['GOOGLE_DATASTORES'] = variables['GOOGLE_DATASTORES'].split()

    if not variables.get('TOKEN', None):
        variables['TOKEN'] = run_cmd(f'gcloud auth print-access-token')

    for k in ["TOKEN", "GOOGLE_BUCKET", 'GOOGLE_DATASTORES']:
        assert variables[k], f"Should have set variable {k}"

    ETLVariables = namedtuple('ETLVariables', PROPERTIES_LIST)
    return ETLVariables(**variables)





#!/usr/bin/env python3

import click
import logging

from anvil.etl.extract import extract
from anvil.etl.transform import transform
from anvil.etl.utility import utility
from anvil.etl.load import load
from anvil.etl import DEFAULT_OUTPUT_PATH, NaturalOrderGroup, read_config
from click_loglevel import LogLevel
from importlib_metadata import distribution

LOG_FORMAT = '%(asctime)s %(name)s %(levelname)-8s %(message)s'
logger = logging.getLogger('etl-cli')


@click.group(cls=NaturalOrderGroup)
@click.option("-l", "--log-level", type=LogLevel(), default=logging.INFO)
@click.option('--output_path', default=DEFAULT_OUTPUT_PATH, help='output path for working files and output.', show_default=True)
@click.option('--config_path', default=None, help='path to config file override.', show_default=True)
@click.pass_context
def cli(ctx, log_level, output_path, config_path):
    """ETL: extract from terra workspaces, google buckets and gen3; transform to FHIR; load to google  Healthcare API."""
    # ensure that ctx.obj exists and is a dict
    # set root logging
    logging.basicConfig(level=log_level, format=LOG_FORMAT)
    ctx.ensure_object(dict)
    ctx.obj['log_level'] = log_level
    ctx.obj['output_path'] = output_path
    ctx.obj['config'] = read_config(config_path)


@cli.command()
def version():
    """Print the version."""
    dist = distribution('pyAnVIL')
    print(dist.version)


# load sub-commands
cli.add_command(extract)
cli.add_command(transform)
cli.add_command(load)
cli.add_command(utility)




"""Test terra wrapper."""

from anvil.terra.api import get_projects, whoami
import logging

logging.getLogger('anvil.test_terra').setLevel(logging.DEBUG)
logger = logging.getLogger('anvil.test_terra')


def test_whoami():
    """Ensure terra."""
    me = whoami()
    logger.debug(me)
    assert me, "MUST have terra identity"


def test_get_projects():
    """Print list of projects."""
    namespace = 'anvil-datastorage'
    project_pattern = 'AnVIL_CCDG.*'
    projects = get_projects([namespace], project_pattern=project_pattern)
    assert len(projects) > 0, "Should have at least 1 project matching {}".format(project_pattern)
    for name in [w['workspace']['name'] for w in projects]:
        assert 'ccdg' in name.lower(), "Only CCDG workspaces should be included"



import os
from collections import defaultdict

import click
import logging
import json

import requests

from anvil.etl.transform import _consortium_from_workspace
from anvil.etl.transformers.fhir_writer import ensure_data_store_name
from anvil.etl.transformers.normalizer import get_pickled_workspace, fetch_workspace_names
from anvil.etl.utilities.shell_helper import run_cmd

logger = logging.getLogger(__name__)


@click.group(name='fhir')
@click.pass_context
def fhir(ctx):
    pass  # dummy



"""Test transformers."""

from anvil.transformers.transformer import Transformer
from anvil.transformers.fhir.transformer import FhirTransformer
from anvil.util.reconciler import Reconciler
import logging

from anvil.terra.workspace import Workspace
from anvil.terra.subject import Subject
from anvil.terra.sample import Sample
from anvil.terra.blob import Blob

import json

logging.getLogger('anvil.test_transformers').setLevel(logging.DEBUG)
logger = logging.getLogger('anvil.test_transformers')


def test_transformer(user_project, namespaces, avro_path, terra_output_path, drs_output_path, project_pattern='AnVIL_CMG_Broad_Muscle_KNC_WGS'):
    """Ensure terra."""
    reconciler = Reconciler('CMG', user_project, namespaces, project_pattern, avro_path, terra_output_path, drs_output_path)
    assert reconciler, "MUST create reconciler"
    assert len(reconciler.workspaces) == 1, "MUST have at least expected number of workspaces"
    for workspace in reconciler.workspaces:
        transformer = Transformer(workspace=workspace)
        items = [item for item in transformer.transform()]
        assert len(items) > 0, "Should produce at least one item."
        assert len([w for w in items if isinstance(w, Workspace)]), "Should have an instance of Workspace"
        assert len([s for s in items if isinstance(s, Subject)]), "Should have an instance of Subject"


def test_emitter(user_project, namespaces, avro_path, terra_output_path, drs_output_path, project_pattern='AnVIL_CMG_Broad_Muscle_KNC_WGS'):
    """Ensure terra."""
    reconciler = Reconciler('CMG', user_project, namespaces, project_pattern, avro_path, terra_output_path, drs_output_path)

    classes = [Workspace, Subject, Sample, Blob]
    emitters = {c.__name__: open(f"/tmp/{c.__name__}.json", "w") for c in classes}

    def get_emitter(item):
        for c in classes:
            if isinstance(item, c):
                return emitters[c.__name__]
        return None

    for workspace in reconciler.workspaces:
        transformer = FhirTransformer(workspace=workspace)
        for item in transformer.transform():
            emitter = get_emitter(item)
            assert emitter, f"Should have an emitter {item}"
            for entity in item.entity():
                json.dump(entity, emitter, separators=(',', ':'))
                emitter.write('\n')

    for emitter in emitters.values():
        emitter.close()

    emitters = {c.__name__: open(f"/tmp/{c.__name__}.json", "r") for c in classes}
    for emitter in emitters.values():
        for line in emitter.readlines():
            entity = json.loads(line)
            assert entity
            # for k, v in entity.items():
            #     assert not isinstance(v, list), f"Should be scalar {k}"
            #     assert not isinstance(v, dict), f"Should be scalar {k}"



import json

import requests
import os
import pytest

from anvil.etl.utilities.shell_helper import run_cmd


@pytest.fixture(scope="session")
def token():
    return os.environ.get('TOKEN', run_cmd("gcloud auth application-default print-access-token"))


@pytest.fixture
def base_api():
    default = 'https://healthcare.googleapis.com/v1beta1/projects/fhir-test-16-342800/locations/us-west2/datasets/anvil-test/fhirStores/dev/fhir'
    return os.environ.get('BASE_API', default)


def test_implementation_guide(token, base_api):

    url = f'{base_api}/ImplementationGuide'

    headers = {
        "Content-Type": "application/fhir+json;charset=utf-8",
        "Authorization": f"Bearer {token}",
    }
    response = requests.get(url, headers=headers)
    response.raise_for_status()
    print(json.dumps(response.json()))
    assert response.json()['total'] > 0, "Should have an IG"
    assert response.json()['entry'][0]['resource']['id'] == 'NCPI-FHIR-Implementation-Guide'


def test_document_reference_invalid_body_no_subject(token, base_api):
    validate_url = f'{base_api}/DocumentReference/$validate'

    headers = {
        "Content-Type": "application/fhir+json;charset=utf-8",
        "Authorization": f"Bearer {token}",
    }

    # missing subject
    invalid_body_no_subject = {
        "resourceType": "DocumentReference",
        "id": "44a58180-e4ff-5a8c-9a1f-db4a76ce6f1f",
        "meta": {
            "profile": [
                "https://ncpi-fhir.github.io/ncpi-fhir-ig/StructureDefinition/ncpi-research-document-reference"
            ]
        },
        "identifier": [
            {
                "system": "https://anvil.terra.bio/#workspaces/anvil-datastorage/1000G-high-coverage-2019",
                "value": "gs://fc-56ac46ea-efc4-4683-b6d5-6d95bed41c5e/CCDG_14151/Project_CCDG_14151_B01_GRM_WGS.cram.2020-02-12/Sample_HG00405/analysis/HG00405.final.cram"
            },
            {
                "system": "urn:ncpi:unique-string",
                "value": "44a58180-e4ff-5a8c-9a1f-db4a76ce6f1f"
            }
        ],
        "status": "current",
        "custodian": {
            "reference": "Organization/1000G-high-coverage-2019"
        },
        "content": [
            {
                "attachment": {
                    "url": "foo://dg.ANV0/a9a4489b-de29-4589-9409-4767860e8e85"
                },
                "format": {
                    "display": "cram"
                }
            }
        ],
        "context": {
            "related": [
                {
                    "reference": "Task/d0767fea-d6f6-5482-be12-9260e6901c4f"
                }
            ]
        }
    }
    response = requests.post(validate_url, json=invalid_body_no_subject, headers=headers)
    response.raise_for_status()
    issues = response.json()['issue']
    errors = [issue['expression'][0] for issue in issues if issue['severity'] == 'error']
    assert 'DocumentReference.subject' in errors, f"Should have raised a 'invalid number of elements, min is 1, got 0' {issues}. To see the expected errors try `java -jar validator_cli.jar  /tmp/invalid_body_no_subject.json -ig ~/client-apis/pyAnVIL/DATA/fhir/IG/` "
    print("invalid_body_no_subject OK")


def test_document_reference_missing_task(token, base_api):
    validate_url = f'{base_api}/DocumentReference/$validate?profile=https://ncpi-fhir.github.io/ncpi-fhir-ig/StructureDefinition/ncpi-drs-document-reference'

    headers = {
        "Content-Type": "application/fhir+json;charset=utf-8",
        "Authorization": f"Bearer {token}",
    }

    # missing task
    invalid_body_no_task = {"resourceType": "DocumentReference", "id": "44a58180-e4ff-5a8c-9a1f-db4a76ce6f1f", "meta": {
        "profile": ["https://ncpi-fhir.github.io/ncpi-fhir-ig/StructureDefinition/ncpi-drs-document-reference"]},
                            "identifier": [{
                                               "system": "https://anvil.terra.bio/#workspaces/anvil-datastorage/1000G-high-coverage-2019",
                                               "value": "gs://fc-56ac46ea-efc4-4683-b6d5-6d95bed41c5e/CCDG_14151/Project_CCDG_14151_B01_GRM_WGS.cram.2020-02-12/Sample_HG00405/analysis/HG00405.final.cram"},
                                           {"system": "urn:ncpi:unique-string",
                                            "value": "44a58180-e4ff-5a8c-9a1f-db4a76ce6f1f"}], "status": "current",
                            "custodian": {"reference": "Organization/1000G-high-coverage-2019"},
                            "subject": {"reference": "Patient/e87f0936-319b-5b86-bc27-3e3fdde77c29"}, "content": [
            {"attachment": {"url": "drs://dg.ANV0/a9a4489b-de29-4589-9409-4767860e8e85"},
             "format": {"display": "cram"}}]}
    response = requests.post(validate_url, json=invalid_body_no_task, headers=headers)
    response.raise_for_status()
    issues = response.json()['issue']
    print(response.json())
    errors = [issue['expression'][0] for issue in issues if issue['severity'] == 'error']
    assert 'DocumentReference.subject' in errors, f"Should have raised a 'invalid number of elements, min is 1, got 0' {issues}. To see the expected errors try `java -jar validator_cli.jar  /tmp/invalid_body_no_subject.json -ig ~/client-apis/pyAnVIL/DATA/fhir/IG/` "
    print("invalid_body_no_task OK")


def test_document_reference_invalid_body_no_custodian(token, base_api):
    validate_url = f'{base_api}/DocumentReference/$validate'

    headers = {
        "Content-Type": "application/fhir+json;charset=utf-8",
        "Authorization": f"Bearer {token}",
    }
    # missing custodian
    invalid_body_no_custodian = {"resourceType": "DocumentReference", "id": "44a58180-e4ff-5a8c-9a1f-db4a76ce6f1f",
                                 "meta": {"profile": [
                                     "http://fhir.ncpi-project-forge.io/StructureDefinition/ncpi-drs-document-reference"]},
                                 "identifier": [{
                                                    "system": "https://anvil.terra.bio/#workspaces/anvil-datastorage/1000G-high-coverage-2019",
                                                    "value": "gs://fc-56ac46ea-efc4-4683-b6d5-6d95bed41c5e/CCDG_14151/Project_CCDG_14151_B01_GRM_WGS.cram.2020-02-12/Sample_HG00405/analysis/HG00405.final.cram"},
                                                {"system": "urn:ncpi:unique-string",
                                                 "value": "44a58180-e4ff-5a8c-9a1f-db4a76ce6f1f"}], "status": "current",
                                 "subject": {"reference": "Patient/e87f0936-319b-5b86-bc27-3e3fdde77c29"}, "content": [
            {"attachment": {"url": "foo://dg.ANV0/a9a4489b-de29-4589-9409-4767860e8e85"},
             "format": {"display": "cram"}}],
                                 "context": {"related": [{"reference": "Task/d0767fea-d6f6-5482-be12-9260e6901c4f"}]}}
    response = requests.post(validate_url, json=invalid_body_no_custodian, headers=headers)
    response.raise_for_status()
    issues = response.json()['issue']
    errors = [issue['expression'][0] for issue in issues if issue['severity'] == 'error']
    assert 'DocumentReference.custodian' in errors, f"Should have raised a 'invalid number of elements, min is 1, got 0' {issues}"


def test_document_reference_valid_body(token, base_api):
    validate_url = f'{base_api}/DocumentReference/$validate'

    headers = {
        "Content-Type": "application/fhir+json;charset=utf-8",
        "Authorization": f"Bearer {token}",
    }

    # all OK
    valid_body = {"resourceType": "DocumentReference", "id": "44a58180-e4ff-5a8c-9a1f-db4a76ce6f1f", "meta": {
        "profile": ["http://fhir.ncpi-project-forge.io/StructureDefinition/ncpi-drs-document-reference"]},
                  "identifier": [
                      {"system": "https://anvil.terra.bio/#workspaces/anvil-datastorage/1000G-high-coverage-2019",
                       "value": "gs://fc-56ac46ea-efc4-4683-b6d5-6d95bed41c5e/CCDG_14151/Project_CCDG_14151_B01_GRM_WGS.cram.2020-02-12/Sample_HG00405/analysis/HG00405.final.cram"},
                      {"system": "urn:ncpi:unique-string", "value": "44a58180-e4ff-5a8c-9a1f-db4a76ce6f1f"}],
                  "status": "current", "custodian": {"reference": "Organization/1000G-high-coverage-2019"},
                  "subject": {"reference": "Patient/e87f0936-319b-5b86-bc27-3e3fdde77c29"}, "content": [
            {"attachment": {"url": "drs://dg.ANV0/a9a4489b-de29-4589-9409-4767860e8e85"},
             "format": {"display": "cram"}}],
                  "context": {"related": [{"reference": "Task/d0767fea-d6f6-5482-be12-9260e6901c4f"}]}}
    response = requests.post(validate_url, json=valid_body, headers=headers)
    response.raise_for_status()
    issues = response.json()['issue']
    errors = [issue for issue in issues if issue['severity'] == 'error']
    assert len(errors) == 0, f"Should not have any errors {errors}"
    # https://github.com/FirelyTeam/firely-net-sdk/issues/1055
    warnings = [issue for issue in issues if issue['severity'] == 'warning' and issue['code'] != 'informational']
    assert len(warnings) == 0, f"Should not have any warnings {warnings}"


def test_document_reference_attachment(token, base_api):
    validate_url = f'{base_api}/DocumentReference/$validate'
    invalid_body_bad_attachment = {"resourceType": "DocumentReference", "id": "44a58180-e4ff-5a8c-9a1f-db4a76ce6f1f",
                                   "meta": {"profile": [
                                       "http://fhir.ncpi-project-forge.io/StructureDefinition/ncpi-drs-document-reference"]},
                                   "identifier": [{
                                                      "system": "https://anvil.terra.bio/#workspaces/anvil-datastorage/1000G-high-coverage-2019",
                                                      "value": "gs://fc-56ac46ea-efc4-4683-b6d5-6d95bed41c5e/CCDG_14151/Project_CCDG_14151_B01_GRM_WGS.cram.2020-02-12/Sample_HG00405/analysis/HG00405.final.cram"},
                                                  {"system": "urn:ncpi:unique-string",
                                                   "value": "44a58180-e4ff-5a8c-9a1f-db4a76ce6f1f"}],
                                   "status": "current",
                                   "custodian": {"reference": "Organization/1000G-high-coverage-2019"},
                                   "subject": {"reference": "Patient/e87f0936-319b-5b86-bc27-3e3fdde77c29"},
                                   "content": [
                                       {"attachment": {"url": "foo://dg.ANV0/a9a4489b-de29-4589-9409-4767860e8e85"},
                                        "format": {"display": "cram"}}],
                                   "context": {"related": [{"reference": "Task/d0767fea-d6f6-5482-be12-9260e6901c4f"}]}}

    headers = {
        "Content-Type": "application/fhir+json;charset=utf-8",
        "Authorization": f"Bearer {token}",
    }

    # bad attachment
    response = requests.post(validate_url, json=invalid_body_bad_attachment, headers=headers)
    response.raise_for_status()
    issues = response.json()['issue']
    errors = [issue['expression'][0] for issue in issues if issue['severity'] == 'error']
    assert 'DocumentReference.attachment' in errors, f"Should have raised an invalid url {issues}\nExpected error for Google FHIR service.  see https://groups.google.com/g/gcp-healthcare-discuss/c/dOKuFXqPlXo for more"


def test_patient(token, base_api):
    validate_url = f'{base_api}/Patient/$validate'

    valid_body = {"resourceType": "Patient", "id": "e87f0936-319b-5b86-bc27-3e3fdde77c29",
                  "meta": {"profile": ["http://hl7.org/fhir/StructureDefinition/Patient"]}, "identifier": [
            {"system": "https://anvil.terra.bio/#workspaces/anvil-datastorage/1000G-high-coverage-2019",
             "value": "ERS4367795"},
            {"system": "urn:ncpi:unique-string", "value": "1000G-high-coverage-2019/Patient/ERS4367795"}],
                  "managingOrganization": {"reference": "Organization/1000G-high-coverage-2019"}}

    headers = {
        "Content-Type": "application/fhir+json;charset=utf-8",
        "Authorization": f"Bearer {token}",
    }

    # all OK
    response = requests.post(validate_url, json=valid_body, headers=headers)
    response.raise_for_status()
    issues = response.json()['issue']
    errors = [issue for issue in issues if issue['severity'] == 'error']
    assert len(errors) == 0, f"Should not have any errors {errors}"
    warnings = [issue for issue in issues if issue['severity'] == 'warning']
    assert len(warnings) == 0, f"Should not have any warnings {warnings}"

    # TODO - if we implement it in schema
    # # missing managingOrganization
    # valid_body_managingOrganization = {"resourceType":"Patient","id":"e87f0936-319b-5b86-bc27-3e3fdde77c29","meta":{"profile":["http://hl7.org/fhir/StructureDefinition/Patient"]},"identifier":[{"system":"https://anvil.terra.bio/#workspaces/anvil-datastorage/1000G-high-coverage-2019","value":"ERS4367795"},{"system":"urn:ncpi:unique-string","value":"1000G-high-coverage-2019/Patient/ERS4367795"}]}
    # response = requests.post(validate_url, json=valid_body_managingOrganization, headers=headers)
    # response.raise_for_status()
    # issues = response.json()['issue']
    # errors = [issue for issue in issues if issue['severity'] == 'error']
    # assert len(errors) > 0, f"Should have at least 1 error {errors}"
    # warnings = [issue for issue in issues if issue['severity'] == 'warning']
    # assert len(warnings) == 0, f"Should not have any warnings {warnings}"


def test_task(token, base_api):
    validate_url = f'{base_api}/Task/$validate'

    headers = {
        "Content-Type": "application/fhir+json;charset=utf-8",
        "Authorization": f"Bearer {token}",
    }

    # all OK
    valid_body = {"resourceType": "Task", "id": "d0767fea-d6f6-5482-be12-9260e6901c4f", "meta": {
        "profile": ["https://ncpi-fhir.github.io/ncpi-fhir-ig/StructureDefinition/ncpi-specimen-task"]}, "identifier": [
        {"system": "https://anvil.terra.bio/#workspaces/anvil-datastorage/1000G-high-coverage-2019",
         "value": "ERS4367795/Task/AnVILInjest"}, {"system": "urn:ncpi:unique-string",
                                                   "value": "1000G-high-coverage-2019/Patient/ERS4367795/Specimen/ERS4367795/Task/AnVILInjest"}],
                  "status": "accepted", "intent": "unknown", "input": [{"type": {"coding": [{"code": "Specimen"}]},
                                                                        "valueReference": {
                                                                            "reference": "Specimen/e87f0936-319b-5b86-bc27-3e3fdde77c29"}}],
                  "output": [{"type": {"coding": [{"code": "DocumentReference"}]}, "valueReference": {
                      "reference": "DocumentReference/44a58180-e4ff-5a8c-9a1f-db4a76ce6f1f"}},
                             {"type": {"coding": [{"code": "DocumentReference"}]}, "valueReference": {
                                 "reference": "DocumentReference/ed123e4e-f8f2-565c-a7f6-c492c9c4ab60"}},
                             {"type": {"coding": [{"code": "DocumentReference"}]}, "valueReference": {
                                 "reference": "DocumentReference/ccb094a2-567b-52dd-92e3-5f16b3b9f7da"}}],
                  "focus": {"reference": "Specimen/e87f0936-319b-5b86-bc27-3e3fdde77c29"},
                  "for": {"reference": "Patient/e87f0936-319b-5b86-bc27-3e3fdde77c29"},
                  "owner": {"reference": "Organization/1000G-high-coverage-2019"}}
    response = requests.post(validate_url, json=valid_body, headers=headers)
    response.raise_for_status()
    issues = response.json()['issue']
    errors = [issue for issue in issues if issue['severity'] == 'error']
    assert len(errors) == 0, f"Should not have any errors {errors}"
    warnings = [issue for issue in issues if issue['severity'] == 'warning']
    assert len(warnings) == 0, f"Should not have any warnings {warnings}"

    # no focus
    invalid_body_no_focus = {"resourceType": "Task", "id": "d0767fea-d6f6-5482-be12-9260e6901c4f", "meta": {
        "profile": ["https://ncpi-fhir.github.io/ncpi-fhir-ig/StructureDefinition/ncpi-specimen-task"]}, "identifier": [
        {"system": "https://anvil.terra.bio/#workspaces/anvil-datastorage/1000G-high-coverage-2019",
         "value": "ERS4367795/Task/AnVILInjest"}, {"system": "urn:ncpi:unique-string",
                                                   "value": "1000G-high-coverage-2019/Patient/ERS4367795/Specimen/ERS4367795/Task/AnVILInjest"}],
                             "status": "accepted", "intent": "unknown", "input": [
            {"type": {"coding": [{"code": "Specimen"}]},
             "valueReference": {"reference": "Specimen/e87f0936-319b-5b86-bc27-3e3fdde77c29"}}], "output": [
            {"type": {"coding": [{"code": "DocumentReference"}]},
             "valueReference": {"reference": "DocumentReference/44a58180-e4ff-5a8c-9a1f-db4a76ce6f1f"}},
            {"type": {"coding": [{"code": "DocumentReference"}]},
             "valueReference": {"reference": "DocumentReference/ed123e4e-f8f2-565c-a7f6-c492c9c4ab60"}},
            {"type": {"coding": [{"code": "DocumentReference"}]},
             "valueReference": {"reference": "DocumentReference/ccb094a2-567b-52dd-92e3-5f16b3b9f7da"}}],
                             "for": {"reference": "Patient/e87f0936-319b-5b86-bc27-3e3fdde77c29"},
                             "owner": {"reference": "Organization/1000G-high-coverage-2019"}}
    response = requests.post(validate_url, json=invalid_body_no_focus, headers=headers)
    response.raise_for_status()
    issues = response.json()['issue']
    errors = [issue['expression'][0] for issue in issues if issue['severity'] == 'error']
    assert 'Task.focus' in errors, f"Should have raised missing focus {issues}"
    warnings = [issue for issue in issues if issue['severity'] == 'warning']
    assert len(warnings) == 0, f"Should not have any warnings {warnings}"

    # no for
    invalid_body_no_for = {"resourceType": "Task", "id": "d0767fea-d6f6-5482-be12-9260e6901c4f", "meta": {
        "profile": ["https://ncpi-fhir.github.io/ncpi-fhir-ig/StructureDefinition/ncpi-specimen-task"]}, "identifier": [
        {"system": "https://anvil.terra.bio/#workspaces/anvil-datastorage/1000G-high-coverage-2019",
         "value": "ERS4367795/Task/AnVILInjest"}, {"system": "urn:ncpi:unique-string",
                                                   "value": "1000G-high-coverage-2019/Patient/ERS4367795/Specimen/ERS4367795/Task/AnVILInjest"}],
                           "status": "accepted", "intent": "unknown", "input": [
            {"type": {"coding": [{"code": "Specimen"}]},
             "valueReference": {"reference": "Specimen/e87f0936-319b-5b86-bc27-3e3fdde77c29"}}], "output": [
            {"type": {"coding": [{"code": "DocumentReference"}]},
             "valueReference": {"reference": "DocumentReference/44a58180-e4ff-5a8c-9a1f-db4a76ce6f1f"}},
            {"type": {"coding": [{"code": "DocumentReference"}]},
             "valueReference": {"reference": "DocumentReference/ed123e4e-f8f2-565c-a7f6-c492c9c4ab60"}},
            {"type": {"coding": [{"code": "DocumentReference"}]},
             "valueReference": {"reference": "DocumentReference/ccb094a2-567b-52dd-92e3-5f16b3b9f7da"}}],
                           "focus": {"reference": "Specimen/e87f0936-319b-5b86-bc27-3e3fdde77c29"},
                           "owner": {"reference": "Organization/1000G-high-coverage-2019"}}
    response = requests.post(validate_url, json=invalid_body_no_for, headers=headers)
    response.raise_for_status()
    issues = response.json()['issue']
    errors = [issue['expression'][0] for issue in issues if issue['severity'] == 'error']
    assert 'Task.for' in errors, f"Should have raised missing focus {issues}"
    warnings = [issue for issue in issues if issue['severity'] == 'warning']
    assert len(warnings) == 0, f"Should not have any warnings {warnings}"

    # no owner
    invalid_body_no_owner = {"resourceType": "Task", "id": "d0767fea-d6f6-5482-be12-9260e6901c4f", "meta": {
        "profile": ["https://ncpi-fhir.github.io/ncpi-fhir-ig/StructureDefinition/ncpi-specimen-task"]}, "identifier": [
        {"system": "https://anvil.terra.bio/#workspaces/anvil-datastorage/1000G-high-coverage-2019",
         "value": "ERS4367795/Task/AnVILInjest"}, {"system": "urn:ncpi:unique-string",
                                                   "value": "1000G-high-coverage-2019/Patient/ERS4367795/Specimen/ERS4367795/Task/AnVILInjest"}],
                             "status": "accepted", "intent": "unknown", "input": [
            {"type": {"coding": [{"code": "Specimen"}]},
             "valueReference": {"reference": "Specimen/e87f0936-319b-5b86-bc27-3e3fdde77c29"}}], "output": [
            {"type": {"coding": [{"code": "DocumentReference"}]},
             "valueReference": {"reference": "DocumentReference/44a58180-e4ff-5a8c-9a1f-db4a76ce6f1f"}},
            {"type": {"coding": [{"code": "DocumentReference"}]},
             "valueReference": {"reference": "DocumentReference/ed123e4e-f8f2-565c-a7f6-c492c9c4ab60"}},
            {"type": {"coding": [{"code": "DocumentReference"}]},
             "valueReference": {"reference": "DocumentReference/ccb094a2-567b-52dd-92e3-5f16b3b9f7da"}}],
                             "focus": {"reference": "Specimen/e87f0936-319b-5b86-bc27-3e3fdde77c29"},
                             "for": {"reference": "Patient/e87f0936-319b-5b86-bc27-3e3fdde77c29"}}
    response = requests.post(validate_url, json=invalid_body_no_owner, headers=headers)
    response.raise_for_status()
    issues = response.json()['issue']
    errors = [issue['expression'][0] for issue in issues if issue['severity'] == 'error']
    assert 'Task.owner' in errors, f"Should have raised missing focus {issues}"
    warnings = [issue for issue in issues if issue['severity'] == 'warning']
    assert len(warnings) == 0, f"Should not have any warnings {warnings}"



"""Provide test fixtures."""

import pytest
import os
from anvil.clients.fhir_client import DispatchingFHIRClient
from anvil.clients.smart_auth import KidsFirstFHIRAuth, GoogleFHIRAuth
from fhirclient.models.researchstudy import ResearchStudy
from fhirclient.models.observation import Observation

@pytest.fixture
def token():
    """Optional access token, if none provided GoogleFHIRAuth will perform `gcloud auth print-access-token` is used."""
    return os.environ.get('TOKEN', None)


@pytest.fixture
def kids_first_cookie():
    """AWSELBAuthSessionCookie cookie captured from https://kf-api-fhir-service.kidsfirstdrc.org browser"""
    assert 'KIDS_FIRST_COOKIE' in os.environ, "AWSELBAuthSessionCookie cookie captured from https://kf-api-fhir-service.kidsfirstdrc.org browser"
    return os.environ['KIDS_FIRST_COOKIE']


@pytest.fixture
def anvil_client(token):
    """Create FHIRClient for AnVIL server."""
    settings = {
        'app_id': __name__,
        'api_bases': [
            'https://healthcare.googleapis.com/v1beta1/projects/fhir-test-16-342800/locations/us-west2/datasets/anvil-test/fhirStores/public/fhir',
        ]
    }
    return DispatchingFHIRClient(settings=settings, auth=GoogleFHIRAuth(access_token=token))


@pytest.fixture
def kids_first_client(kids_first_cookie):
    """Create FHIRClient for Kids First server. 
    :param kids_first_cookie: AWSELBAuthSessionCookie cookie captured from browser after login to https://kf-api-fhir-service.kidsfirstdrc.org"""

    settings = {
        'app_id': __name__,
        'api_bases': ['https://kf-api-fhir-service.kidsfirstdrc.org'],
        'retrieve_all': True
    }
    return DispatchingFHIRClient(settings=settings, auth=KidsFirstFHIRAuth(cookie=kids_first_cookie))


@pytest.fixture
def dbgap_client():
    """Create FHIRClient for dbgap (no auth)."""
    settings = {
        'app_id': __name__,
        'api_bases': ['https://dbgap-api.ncbi.nlm.nih.gov/fhir/x1']
    }
    return DispatchingFHIRClient(settings=settings)


@pytest.fixture
def aggregated_client(token, kids_first_server):
    """Create FHIRClient for AnVIL and Kids First servers."""
    settings = {
        'app_id': __name__,
        'api_bases': [
            'https://healthcare.googleapis.com/v1beta1/projects/fhir-test-11-329119/locations/us-west2/datasets/anvil-test/fhirStores/public/fhir',
        ]
    }
    dispatching_client = DispatchingFHIRClient(settings=settings, auth=GoogleFHIRAuth(access_token=token))
    dispatching_client.clients.append(kids_first_server)
    return dispatching_client


@pytest.fixture
def search_research_studies_with_observations():
    """Search clause for all ResearchStudy with Observation"""
    return ResearchStudy.where(struct={}).include('focus', Observation, reverse=True)


@pytest.fixture
def dbgap_research_studies_with_observations(dbgap_client, search_research_studies_with_observations):
    """
    Perform search on dbgap_server.
    :return: List of resources
    """
    return search_research_studies_with_observations.perform_resources(dbgap_client.server)


@pytest.fixture
def kids_first_research_studies_with_observations(caplog, kids_first_client, search_research_studies_with_observations):
    """
    Perform search on kids_first_server.
    :return: List of resources
    """
    # caplog.set_level(logging.DEBUG)
    return search_research_studies_with_observations.perform_resources(kids_first_client.server)


@pytest.fixture
def anvil_research_studies_with_observations(anvil_client, search_research_studies_with_observations):
    """
    Perform search on anvil_server.
    :return: List of resources
    """
    return search_research_studies_with_observations.perform_resources(anvil_client.server)



"""
These tests request a single record from NCPI servers.
"""

from collections import defaultdict
import os
import pytest

from anvil.clients.fhir_client import DispatchingFHIRClient
from anvil.clients.smart_auth import KidsFirstFHIRAuth, GoogleFHIRAuth

from fhirclient.models.researchstudy import ResearchStudy
from fhirclient.models.observation import Observation
from fhirclient.models.researchsubject import ResearchSubject
from fhirclient.models.bundle import Bundle
from urllib.parse import urlparse

import logging


@pytest.fixture
def token():
    """Optional access token, if none provided GoogleFHIRAuth will perform `gcloud auth print-access-token` is used."""
    return os.environ.get('TOKEN', None)


@pytest.fixture
def kids_first_cookie():
    """AWSELBAuthSessionCookie cookie captured from https://kf-api-fhir-service.kidsfirstdrc.org browser"""
    assert 'KIDS_FIRST_COOKIE' in os.environ
    assert os.environ['KIDS_FIRST_COOKIE'].startswith('AWSELBAuthSessionCookie')
    return os.environ['KIDS_FIRST_COOKIE']


@pytest.fixture
def anvil_server(token):
    """Create FHIRClient for AnVIL server."""
    settings = {
        'app_id': __name__,
        'api_bases': [
            'https://healthcare.googleapis.com/v1beta1/projects/fhir-test-16-342800/locations/us-west2/datasets/anvil-test/fhirStores/public/fhir',
        ]
    }
    return DispatchingFHIRClient(settings=settings, auth=GoogleFHIRAuth(access_token=token))


@pytest.fixture
def kids_first_server(kids_first_cookie):
    """Create FHIRClient for Kids First server. 
    :param kids_first_cookie: AWSELBAuthSessionCookie cookie captured from browser after login to https://kf-api-fhir-service.kidsfirstdrc.org"""

    settings = {
        'app_id': __name__,
        'api_bases': ['https://kf-api-fhir-service.kidsfirstdrc.org']
    }
    return DispatchingFHIRClient(settings=settings, auth=KidsFirstFHIRAuth(cookie=kids_first_cookie))


@pytest.fixture
def dbgap_server():
    """Create FHIRClient for dbgap (no auth)."""
    settings = {
        'app_id': __name__,
        'api_bases': ['https://dbgap-api.ncbi.nlm.nih.gov/fhir/x1']
    }
    return DispatchingFHIRClient(settings=settings)


@pytest.fixture
def query_one():
    """Retrieve a single ResearchStudy."""
    return ResearchStudy.where(struct={'_count': '1'})


def test_simple_kf(kids_first_server, query_one):
    """Return a single resource."""
    resources = query_one.perform_resources(kids_first_server.server)
    assert len(resources) == 1


def test_simple_anvil(anvil_server, query_one):
    """Return a single resource."""
    resources = query_one.perform_resources(anvil_server.server)
    assert len(resources) == 1


def test_simple_dbgap(dbgap_server, query_one):
    """Return a single resource."""
    resources = query_one.perform_resources(dbgap_server.server)
    assert len(resources) == 1



import os
import pytest

from anvil.clients.fhir_client import DispatchingFHIRClient
from anvil.clients.smart_auth import GoogleFHIRAuth
from anvil.etl.utilities.shell_helper import run_cmd
from fhirclient.models.researchstudy import ResearchStudy
import logging


# print("LOG LEVEL:", logging.getLogger().getEffectiveLevel())
# assert logging.getLogger().getEffectiveLevel() == logging.DEBUG, logging.getLogger().getEffectiveLevel()

logger = logging.getLogger(__name__)


@pytest.fixture
def token():
    """Optional access token, if none provided GoogleFHIRAuth will perform `gcloud auth print-access-token` is used."""
    return os.environ.get('TOKEN', None)


@pytest.fixture
def aggregated_server(token):
    """Create FHIRClient for AnVIL and Kids First servers."""
    settings = {
        'app_id': __name__,
        'api_bases': None
    }
    if not token:
        token = run_cmd("gcloud auth application-default print-access-token")
    list_cmd = f'gcloud beta healthcare fhir-stores list --dataset=$GOOGLE_DATASET  --location=$GOOGLE_LOCATION  --format="table[no-heading](ID)"'
    data_stores = [data_store for data_store in  run_cmd(list_cmd).split('\n')]
    settings['api_bases'] = [f'https://healthcare.googleapis.com/v1beta1/projects/fhir-test-16-342800/locations/us-west2/datasets/anvil-test/fhirStores/{data_store}/fhir' for data_store in data_stores]
    logger.info(f"Query dispatching to {len(settings['api_bases'])} data-stores")
    return DispatchingFHIRClient(settings=settings, auth=GoogleFHIRAuth(access_token=token))


@pytest.fixture
def query_all_research_studies():
    """Retrieve a single ResearchStudy."""
    return ResearchStudy.where(struct={'_count': '1000', '_elements': 'id,status'})


@pytest.fixture
def minimum_expected_study_count():
    """How many research studies are there?"""
    return 440


def test_dispatching_query(aggregated_server, query_all_research_studies, minimum_expected_study_count):
    """Retrieve all studies from all data stores."""
    all_studies = query_all_research_studies.perform_resources(aggregated_server.server)
    assert len(all_studies) > minimum_expected_study_count, f"Expected over {minimum_expected_study_count} studies, got {len(all_studies)}"



"""
These tests explore data in dbGAP, AnVIL and Kids First
- they are designed to find corner cases in the way the data was encoded.
- as such they will fail
"""

from collections import defaultdict
import os
import pytest

from anvil.clients.fhir_client import DispatchingFHIRClient
from anvil.clients.smart_auth import KidsFirstFHIRAuth, GoogleFHIRAuth

from fhirclient.models.researchstudy import ResearchStudy
from fhirclient.models.observation import Observation
from fhirclient.models.researchsubject import ResearchSubject
from fhirclient.models.bundle import Bundle
from urllib.parse import urlparse

import logging


@pytest.fixture
def token():
    """Optional access token, if none provided GoogleFHIRAuth will perform `gcloud auth print-access-token` is used."""
    return os.environ.get('TOKEN', None)


@pytest.fixture
def kids_first_cookie():
    """AWSELBAuthSessionCookie cookie captured from https://kf-api-fhir-service.kidsfirstdrc.org browser"""
    assert 'KIDS_FIRST_COOKIE' in os.environ
    assert os.environ['KIDS_FIRST_COOKIE'].startswith('AWSELBAuthSessionCookie')
    return os.environ['KIDS_FIRST_COOKIE']


@pytest.fixture
def anvil_server(token):
    """Create FHIRClient for AnVIL server."""
    settings = {
        'app_id': __name__,
        'api_bases': [
            'https://healthcare.googleapis.com/v1beta1/projects/fhir-test-16-342800/locations/us-west2/datasets/anvil-test/fhirStores/public/fhir',
        ]
    }
    return DispatchingFHIRClient(settings=settings, auth=GoogleFHIRAuth(access_token=token))


@pytest.fixture
def kids_first_server(kids_first_cookie):
    """Create FHIRClient for Kids First server. 
    :param kids_first_cookie: AWSELBAuthSessionCookie cookie captured from browser after login to https://kf-api-fhir-service.kidsfirstdrc.org"""

    settings = {
        'app_id': __name__,
        'api_bases': ['https://kf-api-fhir-service.kidsfirstdrc.org']
    }
    return DispatchingFHIRClient(settings=settings, auth=KidsFirstFHIRAuth(cookie=kids_first_cookie))


@pytest.fixture
def dbgap_server():
    """Create FHIRClient for dbgap (no auth)."""
    settings = {
        'app_id': __name__,
        'api_bases': ['https://dbgap-api.ncbi.nlm.nih.gov/fhir/x1']
    }
    return DispatchingFHIRClient(settings=settings)


@pytest.fixture
def aggregated_server(token, kids_first_server):
    """Create FHIRClient for AnVIL and Kids First servers."""
    settings = {
        'app_id': __name__,
        'api_bases': [
            'https://healthcare.googleapis.com/v1beta1/projects/fhir-test-16-342800/locations/us-west2/datasets/anvil-test/fhirStores/public/fhir',
        ]
    }
    dispatching_client = DispatchingFHIRClient(settings=settings, auth=GoogleFHIRAuth(access_token=token))
    dispatching_client.clients.append(kids_first_server)
    return dispatching_client


@pytest.fixture
def search_research_studies_with_observations():
    """Search clause for all ResearchStudy with Observation"""
    return ResearchStudy.where(struct={}).include('focus', Observation, reverse=True)


@pytest.fixture
def anvil_research_studies_with_observations(anvil_server, search_research_studies_with_observations):
    """
    Perform search on anvil_server.
    :return: List of resources
    """
    return search_research_studies_with_observations.perform_resources(anvil_server.server)


@pytest.fixture
def kids_first_research_studies_with_observations(caplog, kids_first_server, search_research_studies_with_observations):
    """
    Perform search on kids_first_server.
    :return: List of resources
    """
    # caplog.set_level(logging.DEBUG)
    return search_research_studies_with_observations.perform_resources(kids_first_server.server)


@pytest.fixture
def dbgap_research_studies_with_observations(dbgap_server, search_research_studies_with_observations):
    """
    Perform search on dbgap_server.
    :return: List of resources
    """
    return search_research_studies_with_observations.perform_resources(dbgap_server.server)


@pytest.fixture
def all_research_studies_with_observations(aggregated_server, search_research_studies_with_observations):
    """
    Perform search on aggregated_server.
    :return: List of resources
    """
    return search_research_studies_with_observations.perform_resources(aggregated_server.server)


def test_aggregated_server(all_research_studies_with_observations,
                           kids_first_research_studies_with_observations,
                           anvil_research_studies_with_observations):
    """Aggregation should contain resources from both servers"""
    # print([r.relativePath() for r in anvil_research_studies_with_observations])
    assert len(all_research_studies_with_observations) == len(kids_first_research_studies_with_observations) + len(
        anvil_research_studies_with_observations), "Sum of studies should be the same."


def test_ResearchStudy_Observations(all_research_studies_with_observations):
    """All ResearchStudies should have a corresponding Observation"""
    all_observations = {r.relativePath(): r for r in all_research_studies_with_observations if
                        r.__class__.__name__ == 'Observation'}
    all_research_studies = {r.relativePath(): r for r in all_research_studies_with_observations if
                            r.__class__.__name__ == 'ResearchStudy'}
    research_studies_missing_observations = dict(**all_research_studies)
    for observation in all_observations.values():
        del research_studies_missing_observations[observation.focus[0].reference]
    assert len(
        research_studies_missing_observations) == 0, f"ResearchStudies missing Observations {[r.id for r in research_studies_missing_observations.values()]}"


def _validate_ResearchStudy_Condition(research_studies_with_observations):
    """All ResearchStudies should have a coded condition"""
    research_studies_with_observations = {r.relativePath(): r for r in research_studies_with_observations if
                                          r.__class__.__name__ == 'ResearchStudy'}
    research_studies_missing_condition = dict(**research_studies_with_observations)
    for research_study in research_studies_with_observations.values():
        if research_study.condition and len(research_study.condition) > 0:
            # print(f"{research_study.relativePath()} has condition {research_study.condition[0].as_json()}")
            del research_studies_missing_condition[research_study.relativePath()]
    # for k in research_studies_missing_condition:
    #     from pprint import pprint
    #     pprint(research_studies_missing_condition[k].as_json())
    #     break
    assert len(
        research_studies_missing_condition) == 0, f"{len(research_studies_missing_condition)} of {len(research_studies_with_observations)} ResearchStudies missing Condition codeable concept\n{research_studies_missing_condition.keys()}"


def _validate_ResearchStudy_Condition(research_studies_with_observations):
    """All ResearchStudies should have a coded condition"""
    research_studies = {r.relativePath(): r for r in research_studies_with_observations if
                        r.__class__.__name__ == 'ResearchStudy'}
    research_studies_missing_condition = dict(**research_studies)
    for research_study in research_studies.values():
        if research_study.condition and len(research_study.condition) > 0:
            # print(f"{research_study.relativePath()} has condition {research_study.condition[0].as_json()}")
            del research_studies_missing_condition[research_study.relativePath()]
    # for k in research_studies_missing_condition:
    #     from pprint import pprint
    #     pprint(research_studies_missing_condition[k].as_json())
    #     break```
    assert len(
        research_studies_missing_condition) == 0, f"{len(research_studies_missing_condition)} of {len(research_studies)} ResearchStudies missing Condition codeable concept\n{research_studies_missing_condition.keys()}"


def test_AggregatedResearchStudy_Condition(all_research_studies_with_observations):
    """All ResearchStudies should have a coded condition"""
    _validate_ResearchStudy_Condition(all_research_studies_with_observations)


def test_AnvilResearchStudy_Condition(anvil_research_studies_with_observations):
    """All ResearchStudies should have a coded condition"""
    _validate_ResearchStudy_Condition(anvil_research_studies_with_observations)


def test_KidsFirstResearchStudy_Condition(kids_first_research_studies_with_observations):
    """All ResearchStudies should have a coded condition"""
    _validate_ResearchStudy_Condition(kids_first_research_studies_with_observations)


def test_dbGapResearchStudy_Condition(dbgap_research_studies_with_observations):
    """All ResearchStudies should have a coded condition"""
    _validate_ResearchStudy_Condition(dbgap_research_studies_with_observations)


def _validate_ResearchStudy_Focus(research_studies_with_observations):
    """All ResearchStudies should have a focus"""
    research_studies = {r.relativePath(): r for r in research_studies_with_observations if
                        r.__class__.__name__ == 'ResearchStudy'}
    research_studies_missing_focus = dict(**research_studies)
    for research_study in research_studies.values():
        if research_study.focus and len(research_study.focus) > 0:
            # print(f"{research_study.relativePath()} has condition {research_study.condition[0].as_json()}")
            del research_studies_missing_focus[research_study.relativePath()]
    # for k in research_studies_missing_condition:
    #     from pprint import pprint
    #     pprint(research_studies_missing_condition[k].as_json())
    #     break
    assert len(
        research_studies_missing_focus) == 0, f"{len(research_studies_missing_focus)} of {len(research_studies)} ResearchStudies missing Focus codeable concepts"  # \n{research_studies_missing_focus.keys()}


def test_AggregatedResearchStudy_Focus(all_research_studies_with_observations):
    """All ResearchStudies should have a coded Focus"""
    _validate_ResearchStudy_Focus(all_research_studies_with_observations)


def test_AnvilResearchStudy_Focus(anvil_research_studies_with_observations):
    """All ResearchStudies should have a coded Focus"""
    _validate_ResearchStudy_Focus(anvil_research_studies_with_observations)


def test_KidsFirstResearchStudy_Focus(kids_first_research_studies_with_observations):
    """All ResearchStudies should have a coded Focus"""
    _validate_ResearchStudy_Focus(kids_first_research_studies_with_observations)


def test_dbGapResearchStudy_Focus(dbgap_research_studies_with_observations):
    """All ResearchStudies should have a coded Focus"""
    _validate_ResearchStudy_Focus(dbgap_research_studies_with_observations)


def search_research_study_with_subjects(_id):
    """Search clause for all ResearchStudy with Observation"""
    return ResearchStudy.where(struct={'_id': _id}).include('study', ResearchSubject, reverse=True)


def _validate_patient_everything(caplog, fhir_client, research_studies_with_observations):
    assert research_studies_with_observations, 'Should be non-null'
    assert len(research_studies_with_observations) > 0, 'Should have at least one ResearchStudy'
    # pick the first ResearchStudy
    research_studies = {r.relativePath(): r for r in research_studies_with_observations if
                        r.__class__.__name__ == 'ResearchStudy'}
    research_study = list(research_studies.values())[0]
    # get the list of subjects, extract Patient ids
    study_with_subjects = search_research_study_with_subjects(research_study.id).perform_resources(fhir_client.server)
    subjects = {r.relativePath(): r for r in study_with_subjects if r.__class__.__name__ == 'ResearchSubject'}
    patient_ids = [subjects[s].individual.reference.split('/')[-1] for s in subjects]
    patient_bundles = []
    caplog.set_level(logging.DEBUG)
    desired_types = ['ResearchStudy', 'ResearchSubject', 'Patient', 'Specimen', 'Task', 'DocumentReference']
    # KF: ['Patient', 'DocumentReference', 'Specimen', 'ResearchSubject', 'ResearchStudy', 'Practitioner', 'Organization', 'PractitionerRole']
    # Anvil: ['Patient', 'DocumentReference', 'Specimen', 'ResearchSubject']

    for patient_id in patient_ids:
        print(patient_id)
        patient_bundles.append(
            Bundle.read_from(f"Patient/{patient_id}/$everything?_type={','.join(desired_types)}", fhir_client.server))
        break
    resource_counts = {}
    for patient_bundle in patient_bundles:
        resource_counts = defaultdict(int)
        for entry in patient_bundle.entry:
            resource_counts[entry.resource.resource_type] += 1
    for resource_type in desired_types:
        assert resource_type in resource_counts, f"{resource_type} should be in $everything {resource_counts}"
    return resource_counts


def test_kids_first_patient_everything(caplog, kids_first_server, kids_first_research_studies_with_observations):
    """
    Test that a Patient/XXX/$everything returns a set of resources.
    """
    resource_counts = _validate_patient_everything(caplog, kids_first_server,
                                                   kids_first_research_studies_with_observations)
    assert False, resource_counts


def test_anvil_first_patient_everything(caplog, anvil_server, anvil_research_studies_with_observations):
    """
    Test that a Patient/XXX/$everything returns a set of resources.
    """
    research_studies = {r.relativePath(): r for r in anvil_research_studies_with_observations if
                        r.__class__.__name__ == 'ResearchStudy'}
    resource_counts = _validate_patient_everything(caplog, anvil_server,
                                                   [research_studies['ResearchStudy/1000G-high-coverage-2019']])
    assert False, resource_counts


def search_research_subjects_with_patients(study_id):
    """Search clause for all ResearchStudy with Observation"""
    return ResearchSubject.where(struct={'study': f'ResearchStudy/{study_id}'}).include('individual').include('study')


def validate_research_subjects_with_patients(caplog, search, fhir_client):
    """
    Test search on client.server
    """
    from collections import defaultdict
    # caplog.set_level(logging.DEBUG)
    research_subjects_with_patients = search.perform_resources(fhir_client.server)
    print(research_subjects_with_patients)
    resource_types = defaultdict(int)
    for r in research_subjects_with_patients:
        resource_types[r.resource_type] += 1
    assert resource_types['ResearchSubject'] > 1, f"Should have at least one ResearchSubject {resource_types}"
    assert resource_types['ResearchSubject'] == resource_types[
        'Patient'], f"Should have equal number of patients and subjects f{resource_types}"
    for r in research_subjects_with_patients:
        if r.resource_type == 'ResearchStudy':
            print(r.id)
    assert resource_types['ResearchStudy'] > 0, f"Should have at least one ResearchStudy {resource_types}"


def test_dbgap_research_subjects_with_patients(caplog, dbgap_server):
    """
    Test search on dbgap_server.
    """
    validate_research_subjects_with_patients(caplog, search_research_subjects_with_patients('phs002409'), dbgap_server)


def test_kf_research_subjects_with_patients(caplog, kids_first_server):
    """
    Test search on dbgap_server.
    """
    validate_research_subjects_with_patients(caplog, search_research_subjects_with_patients('281300'),
                                             kids_first_server)


def test_anvil_research_subjects_with_patients(caplog, anvil_server):
    """
    Test ResearchStudy search on anvil
    """
    validate_research_subjects_with_patients(caplog, search_research_subjects_with_patients('1000G-high-coverage-2019'),
                                             anvil_server)


def _validateResearchStudy_tag_fullUrl(research_studies_with_observations):
    """Validate .meta.tag has a fullUrl code."""
    research_studies = {r.relativePath(): r for r in research_studies_with_observations if
                        r.__class__.__name__ == 'ResearchStudy'}
    for research_study in research_studies.values():
        has_fullUrl_tag = False
        for tag in research_study.meta.tag:
            if tag['system'] != "https://nih-ncpi.github.io/ncpi-fhir-ig/#fullUrl":
                continue
            url_parts = urlparse(tag['code'])
            assert all([url_parts.scheme, url_parts.netloc]), f"{tag} code should be a valid url"
            has_fullUrl_tag = True
        assert has_fullUrl_tag, f"Should contain fullURL tag {research_study.meta.tags}"


def test_dbGapResearchStudy_tag_fullUrl(dbgap_research_studies_with_observations):
    """All ResearchStudies should have a valid source"""
    _validateResearchStudy_tag_fullUrl(dbgap_research_studies_with_observations)


def test_AnvilResearchStudy_tag_fullUrl(anvil_research_studies_with_observations):
    """All ResearchStudies should have a valid source"""
    _validateResearchStudy_tag_fullUrl(anvil_research_studies_with_observations)


def test_KidsFirstResearchStudy_tag_fullUrl(kids_first_research_studies_with_observations):
    """All ResearchStudies should have a valid source"""
    _validateResearchStudy_tag_fullUrl(kids_first_research_studies_with_observations)


def test_AllResearchStudy_tag_fullUrl(all_research_studies_with_observations):
    """All ResearchStudies should have a url source, used to disambiguate combined results."""
    research_studies = [r for r in all_research_studies_with_observations if r.__class__.__name__ == 'ResearchStudy']
    sources = defaultdict(int)
    for research_study in research_studies:
        url_parts = urlparse(research_study.meta.source)
        assert all([url_parts.scheme, url_parts.netloc]), f"{research_study.meta.source} should be a valid url"
        sources[url_parts.netloc] += 1
    assert len(sources) > 1, f"Should have more than one source in a combined result {sources}"


def test_simple_kf(kids_first_server):
    studies = ResearchStudy.where(struct={'_count': '1'}).perform_resources(kids_first_server.server)
    assert len(studies) == 1


def test_simple_anvil(anvil_server):
    studies = ResearchStudy.where(struct={'_count': '1'}).perform_resources(anvil_server.server)
    assert len(studies) == 1


def test_simple_dbgap(dbgap_server):
    studies = ResearchStudy.where(struct={'_count': '1'}).perform_resources(dbgap_server.server)
    assert len(studies) == 1



from .bioontology_lookup import lookup_term

raw_terms = """
    Adipose - Subcutaneous
    Adipose - Visceral (Omentum)
    Adrenal Gland
    Artery - Aorta
    Artery - Coronary
    Artery - Tibial
    Bladder
    Brain - Amygdala
    Brain - Anterior cingulate cortex (BA24)
    Brain - Caudate (basal ganglia)
    Brain - Cerebellar Hemisphere
    Brain - Cerebellum
    Brain - Cortex
    Brain - Frontal Cortex (BA9)
    Brain - Hippocampus
    Brain - Hypothalamus
    Brain - Nucleus accumbens (basal ganglia)
    Brain - Putamen (basal ganglia)
    Brain - Spinal cord (cervical c-1)
    Brain - Substantia nigra
    Breast - Mammary Tissue
    Cells - Cultured fibroblasts
    Cells - EBV-transformed lymphocytes
    Cervix - Ectocervix
    Cervix - Endocervix
    Colon - Sigmoid
    Colon - Transverse
    Esophagus - Gastroesophageal Junction
    Esophagus - Mucosa
    Esophagus - Muscularis
    Fallopian Tube
    Heart - Atrial Appendage
    Heart - Left Ventricle
    Kidney - Cortex
    Kidney - Medulla
    Liver
    Lung
    Minor Salivary Gland
    Muscle - Skeletal
    Nerve - Tibial
    Ovary
    Pancreas
    Pituitary
    Prostate
    Skin - Not Sun Exposed (Suprapubic)
    Skin - Sun Exposed (Lower leg)
    Small Intestine - Terminal Ileum
    Spleen
    Stomach
    Testis
    Thyroid
    Uterus
    Vagina
    Whole Blood
""".split("\n")

ontology_text = {'http://github.com/obophenotype/uberon/UBERON_0002190': 'Adipose - Subcutaneous',
                 'http://github.com/obophenotype/uberon/UBERON_0014454': 'Adipose - Visceral (Omentum)',
                 'http://github.com/obophenotype/uberon/UBERON_0002369': 'Adrenal Gland',
                 'http://github.com/obophenotype/uberon/UBERON_0012254': 'Artery - Aorta',
                 'http://github.com/obophenotype/uberon/UBERON_0001621': 'Artery - Coronary',
                 'http://github.com/obophenotype/uberon/UBERON_0007610': 'Artery - Tibial',
                 'http://github.com/obophenotype/uberon/UBERON_0018707': 'Bladder',
                 'http://github.com/obophenotype/uberon/UBERON_0001876': 'Brain - Amygdala',
                 'http://github.com/obophenotype/uberon/UBERON_0009835': 'Brain - Anterior cingulate cortex (BA24)',
                 'http://github.com/obophenotype/uberon/UBERON_0010011': 'Brain - Putamen (basal ganglia)',
                 'http://github.com/obophenotype/uberon/UBERON_0002245': 'Brain - Cerebellar Hemisphere',
                 'http://github.com/obophenotype/uberon/UBERON_0002037': 'Brain - Cerebellum',
                 'http://github.com/obophenotype/uberon/UBERON_0001851': 'Brain - Cortex',
                 'http://github.com/obophenotype/uberon/UBERON_0001870': 'Brain - Frontal Cortex (BA9)',
                 'http://github.com/obophenotype/uberon/UBERON_0001898': 'Brain - Hypothalamus',
                 'http://github.com/obophenotype/uberon/UBERON_0001882': 'Brain - Nucleus accumbens (basal ganglia)',
                 'http://github.com/obophenotype/uberon/UBERON_0002726': 'Brain - Spinal cord (cervical c-1)',
                 'http://github.com/obophenotype/uberon/UBERON_0002038': 'Brain - Substantia nigra',
                 'http://github.com/obophenotype/uberon/UBERON_0003584': 'Breast - Mammary Tissue',
                 'http://github.com/obophenotype/uberon/UBERON_0006591': 'Cells - EBV-transformed lymphocytes',
                 'http://github.com/obophenotype/uberon/UBERON_0012249': 'Cervix - Ectocervix',
                 'http://github.com/obophenotype/uberon/UBERON_0000458': 'Cervix - Endocervix',
                 'http://github.com/obophenotype/uberon/UBERON_0001159': 'Colon - Sigmoid',
                 'http://github.com/obophenotype/uberon/UBERON_0001157': 'Colon - Transverse',
                 'http://github.com/obophenotype/uberon/UBERON_0002469': 'Esophagus - Mucosa',
                 'http://github.com/obophenotype/uberon/UBERON_0004648': 'Esophagus - Muscularis',
                 'http://github.com/obophenotype/uberon/UBERON_0003889': 'Fallopian Tube',
                 'http://github.com/obophenotype/uberon/UBERON_0002084': 'Heart - Left Ventricle',
                 'http://github.com/obophenotype/uberon/UBERON_0001225': 'Kidney - Cortex',
                 'http://github.com/obophenotype/uberon/UBERON_0001294': 'Kidney - Medulla',
                 'http://github.com/obophenotype/uberon/UBERON_0002107': 'Liver',
                 'http://github.com/obophenotype/uberon/UBERON_0002048': 'Lung',
                 'http://github.com/obophenotype/uberon/UBERON_0001830': 'Minor Salivary Gland',
                 'http://github.com/obophenotype/uberon/UBERON_0001323': 'Nerve - Tibial',
                 'http://github.com/obophenotype/uberon/UBERON_0000992': 'Ovary',
                 'http://github.com/obophenotype/uberon/UBERON_0001264': 'Pancreas',
                 'http://github.com/obophenotype/uberon/UBERON_0000007': 'Pituitary',
                 'http://github.com/obophenotype/uberon/UBERON_0002367': 'Prostate',
                 'http://github.com/obophenotype/uberon/UBERON_0036149': 'Skin - Not Sun Exposed (Suprapubic)',
                 'http://github.com/obophenotype/uberon/UBERON_0004264': 'Skin - Sun Exposed (Lower leg)',
                 'http://github.com/obophenotype/uberon/UBERON_0002108': 'Small Intestine - Terminal Ileum',
                 'http://github.com/obophenotype/uberon/UBERON_0002106': 'Spleen',
                 'http://github.com/obophenotype/uberon/UBERON_0000945': 'Stomach',
                 'http://github.com/obophenotype/uberon/UBERON_0000473': 'Testis',
                 'http://github.com/obophenotype/uberon/UBERON_0002046': 'Thyroid',
                 'http://github.com/obophenotype/uberon/UBERON_0000995': 'Uterus',
                 'http://github.com/obophenotype/uberon/UBERON_0000996': 'Vagina',
                 'http://github.com/obophenotype/uberon/UBERON_0000178': 'Whole Blood'}

# manual lookup
ontology_text['http://github.com/obophenotype/uberon/UBERON_0004857'] = 'Muscle - Skeletal'
ontology_text['http://github.com/obophenotype/uberon/UBERON_0006618'] = 'Heart - Atrial Appendage'
ontology_text['http://github.com/obophenotype/uberon/UBERON_0007650'] = 'Esophagus - Gastroesophageal Junction'
ontology_text['http://www.ebi.ac.uk/efo/EFO_0002009'] = 'Cells - Cultured fibroblasts'
ontology_text['http://github.com/obophenotype/uberon/UBERON_0001873'] = 'Brain - Caudate (basal ganglia)'
ontology_text['http://github.com/obophenotype/uberon/UBERON_0001954'] = 'Brain - Hippocampus'

text_ontology = {v: k for k, v in ontology_text.items()}


def lookup_body_site(term) -> (str, str):
    """Return (system, code)."""
    assert term in text_ontology, f"Could not find body_site: >{term}<"
    system_parts = text_ontology[term].split('/')
    return '/'.join(system_parts[:-1]), system_parts[-1]


def lookup_raw_terms():
    """Lookup raw terms."""
    ontology_text_ = {}
    raw_terms = [t.strip(' ') for t in raw_terms if len(t) > 1]

    for term in raw_terms:
        items = [item for ontology_preference, item in lookup_term(term, ontology_preferences=['UBERON'])]
        if len(items) == 0:
            continue
        item = items[0]
        ontology_text_[item['@id']] = term

    print(ontology_text_)


if __name__ == "__main__":
   lookup_raw_terms()


"""Instances of this class handle authorizing and talking to Google Healthcare API FHIR Service."""


import logging
import concurrent.futures
from urllib.parse import urlparse
from fhirclient import client, server as fhirclient_server
from fhirclient.models.meta import Meta
from fhirclient.models.bundle import Bundle
import resource

from fhirclient.server import FHIRPermissionDeniedException
import fhirclient.models.identifier as FHIRIdentifier

logger = logging.getLogger(__name__)


class FHIRClient(client.FHIRClient):
    """Instances of this class handle authorizing and talking to Google Healthcare API FHIR Service.

    Parameters:
        See https://github.com/smart-on-fhir/client-py/blob/master/fhirclient/client.py#L19

    Returns:
        Instance of client, with injected authorization method

    Examples: ::
        from anvil.fhir.client import FHIRClient
        settings = {
            'app_id': 'my_web_app',
            'api_base': 'https://healthcare.googleapis.com/v1/projects/gcp-testing-308520/locations/us-east4/datasets/testset/fhirStores/fhirstore/fhir'
        }
        smart = FHIRClient(settings=settings)
        assert smart.ready, "server should be ready"
        # search for all ResearchStudy
        import fhirclient.models.researchstudy as rs
        [s.title for s in rs.ResearchStudy.where(struct={}).perform_resources(smart.server)]
        >>>
        ['1000g-high-coverage-2019', 'my NCPI research study example']


    """

    def __init__(self, *args, **kwargs):
        """Pass args to super, add authenticator, prepares connection."""
        # grab auth if passed
        auth = None
        if 'auth' in kwargs:
            auth = kwargs['auth']
            del kwargs['auth']
        super(FHIRClient, self).__init__(*args, **kwargs)
        client_major_version = int(client.__version__.split('.')[0])
        assert client_major_version >= 4, f"requires version >= 4.0.0 current version {client.__version__} `pip install -e git+https://github.com/smart-on-fhir/client-py#egg=fhirclient`"
        if auth:
            logger.debug("Setting auth")
            self.server.auth = auth
            self.server.session.hooks['response'].append(self.server.auth.handle_401)
        self.prepare()
        assert self.ready, "server should be ready"


class DispatchingFHIRClient(client.FHIRClient):
    """We monkey patch Search.perform, dispatching worker threads for each endpoint in api_bases.
    Same auth handler used for all connections.

    Parameters:
        See https://github.com/smart-on-fhir/client-py/blob/master/fhirclient/client.py#L19

    :param settings.api_bases: The servers against which to perform the search **settings.api_base ignored**
    :param auth: An instance of FHIRAuth that will authenticate each request.
    :param max_workers: Number of thread workers, if null will be set to the maximum of open files per process
    or the number of api bases, whichever is less.

    Returns:
        Instance of client, with injected authorization method

    Examples: ::
        from anvil.clients.fhir_client import DispatchingFHIRClient
        from fhirclient.models.researchstudy import ResearchStudy
        from collections import defaultdict
        from pprint import pprint

        settings = {
            'app_id': 'my_web_app',
            'api_bases': [
                'https://healthcare.googleapis.com/v1beta1/projects/fhir-test-11-329119/locations/us-west2/datasets/anvil-test/fhirStores/public/fhir',
                'https://healthcare.googleapis.com/v1beta1/projects/fhir-test-11-329119/locations/us-west2/datasets/anvil-test/fhirStores/pending/fhir',
            ]
        }
        smart = DispatchingFHIRClient(settings=settings, auth=GoogleFHIRAuth())

        # search for all ResearchStudy, index by source
        studies = defaultdict(list)
        for s in ResearchStudy.where(struct={'_count':'1000'}).perform_resources(smart.server):
            studies[s.meta.source].append(s)

        pprint({k: len(v) for k,v in studies.items()})
        >>>
        {'https://healthcare.googleapis.com/v1beta1/projects/fhir-test-11-329119/locations/us-west2/datasets/anvil-test/fhirStores/pending/fhir/': 259,
        'https://healthcare.googleapis.com/v1beta1/projects/fhir-test-11-329119/locations/us-west2/datasets/anvil-test/fhirStores/public/fhir/': 393}

    """

    def __init__(self, *args, **kwargs):
        """Pass args to super, patches `perform` to our dispatching version."""
        # use the first entry as 'our' server
        _settings = dict(kwargs['settings'])
        api_base = _settings['api_bases'].pop()
        _settings['api_base'] = api_base
        kwargs['settings'] = _settings

        # grab retrieve_all if passed
        self._retrieve_all = False
        if 'retrieve_all' in kwargs['settings']:
            self._retrieve_all = kwargs['settings']['retrieve_all']
            del kwargs['settings']['retrieve_all']

        # grab auth if passed
        auth = None
        if 'auth' in kwargs:
            auth = kwargs['auth']
            del kwargs['auth']

        # determine number of worker threads
        max_workers = None

        if 'max_workers' in kwargs:
            max_workers = kwargs['max_workers']
            del kwargs['max_workers']

        if not max_workers:
            max_workers = int(resource.getrlimit(resource.RLIMIT_NOFILE)[0]/3)
            if max_workers > len(_settings['api_bases']):
                max_workers = max(len(_settings['api_bases']), 1)
            logger.debug(f"Setting number of threads to {max_workers}")

        self._max_workers = max_workers

        # normal setup with our authenticator
        super(DispatchingFHIRClient, self).__init__(*args, **kwargs)
        client_major_version = int(client.__version__.split('.')[0])
        assert client_major_version >= 4, f"requires version >= 4.0.0 current version {client.__version__} `pip install -e git+https://github.com/smart-on-fhir/client-py#egg=fhirclient`"

        if auth:
            self.server.auth = auth
            self.server.session.hooks['response'].append(self.server.auth.handle_401)

        self.prepare()
        assert self.ready, "server should be ready"

        # set up an array of FHIRClients, including this instance, in self._clients
        # re-use authenticator
        self._clients = [self]
        self._api_bases = _settings['api_bases']
        for api_base in self._api_bases:
            __settings = dict(_settings)
            __settings['api_base'] = api_base
            _client = client.FHIRClient(settings=__settings)
            _client.server.auth = self.server.auth
            _client.server.session.hooks['response'].append(self.server.auth.handle_401)
            _client.prepare()
            self._clients.append(_client)

        # monkey patch Search.perform if we haven't already
        from fhirclient.models.fhirsearch import FHIRSearch
        if not hasattr(FHIRSearch, '_anvil_patch'):
            FHIRSearch._anvil_patch = True
            original_perform = FHIRSearch.perform
            me = self

            def _perform(self, server):
                """Dispatch query to api_bases."""
                # FHIRSearch can be used by multiple classes, don't dispatch unless one of ours
                if not server.client.__class__.__name__ == 'DispatchingFHIRClient':
                    logger.debug(f"* * * * * * * original_perform {server.client.__class__.__name__}")
                    return original_perform(self, server)

                def _worker(self, server):
                    """Dispatches request to underlying class, return an entry indexed by base uri.

                    Sets bundle.meta.source
                    See https://www.hl7.org/fhir/resource-definitions.html#Meta.source

                    :param server: The server against which to perform the search
                    :_results: Result of operation added to this array
                    """
                    logger.debug(f"worker starting {server.base_uri}")
                    _worker_results = []
                    _result = original_perform(self, server)
                    logger.debug(f"worker got {_result}")

                    while _result:

                        # add source to meta if it doesn't already exist
                        if not _result.meta:
                            _result.meta = Meta()
                        if not _result.meta.source:
                            _result.meta.source = server.base_uri
                        _worker_results.append(_result)
                        if not me._retrieve_all:
                            break

                        # follow `next` link for pagination
                        if hasattr(_result, 'link'):
                            _next = next((lnk.as_json() for lnk in _result.link if lnk.relation == 'next'), None)
                            _result = None
                            if _next:
                                logger.debug(f"has next {_next}")
                                # request_json takes a full path & query (not host)
                                parts = urlparse(_next['url'])
                                assert len(parts.query) > 0, parts
                                path = f"{parts.path}?{parts.query}"
                                logger.debug(f"attempting next {path}")
                                res = server.request_json(path)
                                _result = Bundle(res)
                                _result.origin_server = server
                        else:
                            _result = None
                    logger.debug(f"worker done {len(_worker_results)}")
                    return _worker_results

                logger.debug("starting threads")
                results = []
                # We can use a with statement to ensure threads are cleaned up promptly
                with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
                    # Start the load operations and mark each future with its URL
                    future_result = {executor.submit(_worker, self, __client.server, ): __client for __client
                                     in server.client._clients}
                    for future in concurrent.futures.as_completed(future_result):
                        try:
                            worker_results = future.result()
                            results.extend(worker_results)
                        except Exception as exc:
                            if 'FHIRPermissionDeniedException' in exc.__class__.__name__:
                                # requests.response embedded in exception
                                logger.error(f"{str(exc)} {exc.args[0].url}")
                            raise exc
                return results

            # monkey patch
            FHIRSearch.perform = _perform

            # since perform returns an array, patch _perform_resources as well.
            def _perform_resources(self, server):
                """Perform the search by calling `perform`, then extracts all Bundle entries and returns a list of Resource instances.

                Sets resource.meta.source
                See https://www.hl7.org/fhir/resource-definitions.html#Meta.source

                :param server: The server against which to perform the search
                :returns: A list of Resource instances
                """
                # flatten into an array of resources
                bundles = self.perform(server)
                resources = []
                if bundles is not None:
                    if not isinstance(bundles, list):
                        bundles = [bundles]
                    for bundle in bundles:
                        if bundle.entry:
                            for entry in bundle.entry:
                                if not entry.resource.meta:
                                    entry.resource.meta = Meta()

                                # if not entry.resource.meta.source:
                                #     assert bundle.meta, bundle.as_json()
                                #     entry.resource.meta.source = bundle.meta.source

                                # add tag for fullURL, allows caller
                                # to disambiguate resources returned from different base URL
                                if not entry.resource.meta.tag:
                                    entry.resource.meta.tag = []
                                entry.resource.meta.tag.append(
                                    FHIRIdentifier.Identifier(
                                        {
                                            "system": "https://nih-ncpi.github.io/ncpi-fhir-ig/#fullUrl",
                                            "value": entry.fullUrl
                                        }
                                    )
                                )
                                resources.append(entry.resource)
                return resources
            FHIRSearch.perform_resources = _perform_resources
            logger.debug("Patched FHIRSearch")

    @property
    def clients(self):
        """Expose list of clients instantiated from settings.api_bases."""
        return self._clients

    def dispatch(self, _worker, *args, **kwargs):
        """Execute callback(server) on thread for every server. """
        logger.debug("starting threads")
        results = []
        # We can use a with statement to ensure threads are cleaned up promptly
        with concurrent.futures.ThreadPoolExecutor(max_workers=self._max_workers) as executor:
            # Start the load operations and mark each future with its URL
            future_result = {executor.submit(_worker, _client.server, *args, **kwargs, ): _client for _client in self._clients}
            for future in concurrent.futures.as_completed(future_result):
                try:
                    worker_results = future.result()
                    results.append(worker_results)
                except Exception as exc:
                    if 'FHIRPermissionDeniedException' in exc.__class__.__name__:
                        # requests.response embedded in exception
                        logger.error(f"{str(exc)} {exc.args[0].url}")
                    logger.error(f"{str(exc)} {exc.args[0].url}")
                    raise exc
        return results




"""Test gen3_auth."""
import os

import pytest

from anvil.clients.gen3_auth import Gen3TerraAuth
from anvil.clients.gen3_auth import TERRA_TOKEN_URL
from gen3.submission import Gen3Submission
from gen3.auth import Gen3Auth

import logging
logging.getLogger('anvil.test_gen3_auth').setLevel(logging.DEBUG)
logger = logging.getLogger('anvil.test_gen3_auth')


@pytest.fixture
def submission_client(terra_auth_url, user_email, gen3_endpoint):
    """Query terra_auth_url for access token and ensure inserted onto all gen3 requests."""
    auth = Gen3TerraAuth(endpoint=gen3_endpoint, terra_auth_url=terra_auth_url, user_email=user_email)
    return Gen3Submission(gen3_endpoint, auth)


@pytest.fixture
def native_submission_client(gen3_endpoint):
    """Query terra_auth_url for access token and ensure inserted onto all gen3 requests."""
    # Install n API Key downloaded from the
    # commons' "Profile" page at ~/.gen3/credentials.json
    path = os.path.expanduser("~/.gen3/credentials.json")
    if not os.path.isfile(path):
        logger.warning(f"No gen3 credentials file found at {path}")
    else:
        from datetime import datetime, timezone
        from pathlib import Path
        stat_result = Path(path).stat()
        modified = datetime.fromtimestamp(stat_result.st_mtime, tz=timezone.utc)
        now = datetime.now(tz=timezone.utc)
        delta = now - modified
        if delta.days > 30:
            logger.warning(f"Gen3 credentials file found at {path} are over 30 days old")
    auth = Gen3Auth(endpoint=gen3_endpoint)
    return Gen3Submission(gen3_endpoint, auth)


def test_get_graphql_schema(submission_client):
    """Read schema (AFAIK, gen3 does not check auth)."""
    schema = submission_client.get_graphql_schema()
    assert schema, "MUST be able to access open url"
    # save for downstream development
    # with open("schema.json", "w") as outs:
    #     json.dump(schema, outs)


def test_get_programs(submission_client):
    """Read programs list (AFAIK, gen3 does not check auth)."""
    programs = submission_client.get_programs()
    assert len(programs) > 0, 'MUST have at least one program'


def test_get_projects(submission_client):
    """Read project list."""
    programs = submission_client.get_programs()
    # >>> {'links': ['/v0/submission/open_access', '/v0/submission/CF']}
    programs = [p.split('/')[-1] for p in programs['links']]
    for program in programs:
        projects = submission_client.get_projects(program)
        assert len(projects) > 0, f'MUST have at least one project {program}'
        # >>> {'links': ['/v0/submission/open_access/1000Genomes']}
        # >>> {'links': ['/v0/submission/CF/GTEx']}


def test_query(caplog, submission_client):
    caplog.set_level(logging.DEBUG)
    """Submit simple graphql query."""
    query = '{project(first:0) {code,  subjects {submitter_id}, programs {name}  }}'
    results = submission_client.query(query)
    logger.debug(f"graphql {results}")
    assert len(results['data']['project']) > 1, "MUST have more than one project"
    for project in results['data']['project']:
        if len(project['subjects']) == 0:
            logger.warning(f"gen3 {project['code']} SHOULD have subjects?")
    program_names = set()
    for project in results['data']['project']:
        for program in project['programs']:
            program_names.add(program['name'])
    assert len(program_names) > 1, "MUST have more than one program"


@pytest.mark.skip(reason="TODO: no manifests returned?")
def test_get_project_manifest(submission_client):
    """Read program/project list, request manifest for each."""
    programs = submission_client.get_programs()
    # >>> {'links': ['/v0/submission/open_access', '/v0/submission/CF']}
    programs = [p.split('/')[-1] for p in programs['links']]
    failed_manifests = []
    for program in programs:
        projects = submission_client.get_projects(program)
        # >>> {'links': ['/v0/submission/open_access/1000Genomes']}
        # >>> {'links': ['/v0/submission/CF/GTEx']}
        for project in projects['links']:
            project_code = project.split('/')[-1]
            logger.debug(f"parameters to get_project_manifest: program {program}, project_code {project_code}")
            project_manifest = submission_client.get_project_manifest(program, project_code)
            # assert project_manifest, f"MUST have manifest {program}/{project_code}"
            if not project_manifest:
                failed_manifests.append(f"MUST have manifest {program}/{project_code}")
            logger.debug(project_manifest)
    assert len(failed_manifests) == 0, failed_manifests


def test_get_project_dictionary(submission_client):
    """Read program/project list, request dictionary for each."""
    programs = submission_client.get_programs()
    # >>> {'links': ['/v0/submission/open_access', '/v0/submission/CF']}
    programs = [p.split('/')[-1] for p in programs['links']]
    for program in programs:
        projects = submission_client.get_projects(program)
        # >>> {'links': ['/v0/submission/open_access/1000Genomes']}
        # >>> {'links': ['/v0/submission/CF/GTEx']}
        for project in projects['links']:
            project_code = project.split('/')[-1]
            logger.debug(f"parameters to get_project_dictionary: program {program}, project_code {project_code}")
            project_dictionary = submission_client.get_project_dictionary(program, project_code)
            assert project_dictionary, f"MUST have manifest {program}/{project_code}"


def test_get_dictionary_all(submission_client):
    """Read program/project list, requests dictionary for each."""
    dictionary_all = submission_client.get_dictionary_all()
    logger.debug(dictionary_all.keys())
    for k in ['core_metadata_collection', 'data_release', 'discovery', 'family', 'metaschema', 'program', 'project', 'root', 'sample', 'sequencing', 'subject']:
        assert k in dictionary_all, f'dictionary MUST contain {k}'


def test_project_access(native_submission_client, submission_client):
    """Native token and token proxied by terra should return same project list."""
    query = """{
        project(first: 1000) {
            code
        }
    }"""
    proxy_token_results = submission_client.query(query)
    logger.debug(f"graphql {proxy_token_results}")
    native_results = native_submission_client.query(query)
    logger.debug(f"graphql {native_results}")
    assert set([p['code'] for p in proxy_token_results['data']['project']]) == set([p['code'] for p in native_results['data']['project']]), "MUST return same set of projects,  You may need to re-link your account."




from fhirclient.models.condition import Condition
from fhirclient.models.researchstudy import ResearchStudy
from collections import defaultdict
import pytest
import json
from os.path import exists

EXPECTED_CONDITION_CODE_SYSTEMS = ['http://purl.obolibrary.org/obo/hp.owl', 'http://purl.obolibrary.org/obo/mondo.owl', 'http://snomed.info/sct', 'http://purl.obolibrary.org/obo/ncit.owl', 'https://www.who.int/classifications/classification-of-diseases']

def _conditions(client):
    """
    Perform search on client's server .
    :return: List of Conditions
    """
    return Condition.where(struct={'_count':'1000'}).perform_resources(client.server)


@pytest.fixture
def kids_first_conditions(kids_first_client):
    """
    Perform search on kids_first .
    :return: List of Conditions
    """
    if exists("kids_first_conditions.json"):
        with open("kids_first_conditions.json", "r") as input:
            conditions = []
            for line in input:
                conditions.append(Condition(json.loads(line)))
            return conditions
    conditions = _conditions(kids_first_client)
    with open("kids_first_conditions.json", "w") as output:
        for condition in kids_first_conditions:
            json.dump(condition.as_json(), output, separators=(',', ':'))
            output.write("\n")
    return conditions


def _validate_kidsfirst_conditions(kids_first_conditions):
    """There should be many Conditions."""
    assert kids_first_conditions, "Should return conditions"
    assert len(kids_first_conditions) > 0, "Should return conditions"


# def test_kidsfirst_conditions(kids_first_conditions):
#     """There should be many Conditions."""
#     _validate_kidsfirst_conditions(kids_first_conditions)


# def test_kidsfirst_condition_missing_coding(kids_first_conditions):
#     """All conditions should have a coding."""
#     _validate_kidsfirst_conditions(kids_first_conditions)
#     missing_coding = defaultdict(int)
#     for condition in kids_first_conditions:
#         if not condition.code.coding:
#             missing_coding[f"Condition/{condition.id}"] += 1
#     assert len(missing_coding) == 0, f"All conditions should have a coding\n{missing_coding}"


# def test_kidsfirst_condition_systems(kids_first_conditions):
#     """There should be many a limited set of systems."""
#     _validate_kidsfirst_conditions(kids_first_conditions)
#     systems = defaultdict(int) 
#     for condition in kids_first_conditions:
#         if not condition.code.coding:
#             continue
#         for coding in condition.code.coding:
#             systems[coding.system] += 1
#         assert coding.system in EXPECTED_CONDITION_CODE_SYSTEMS, f"{coding.system} is an unexpected code system" 


# def test_kidsfirst_condition_systems(kids_first_conditions):
#     """There should be many a limited set of systems."""
#     _validate_kidsfirst_conditions(kids_first_conditions)
#     systems = defaultdict(int) 

#     systems = defaultdict(int) 
#     codes = defaultdict(int)
#     prefixes = defaultdict(int)
#     missing_coding = defaultdict(int)
#     for condition in kids_first_conditions:
#         if not condition.code.coding:
#             missing_coding[f"Condition/{condition.id}"] += 1
#             continue
#         for coding in condition.code.coding:
#             codes[coding.code] += 1
#             systems[coding.system] += 1
#             prefixes[coding.code.split(':')[0]] += 1
#     print(systems)
#     print(codes)
#     print(prefixes)
#     print(missing_coding)


@pytest.fixture
def dbgap_conditions(dbgap_client):
    """
    Perform search on kids_first .
    :return: List of Conditions
    """
    if exists("dbgap_conditions.json"):
        with open("dbgap_conditions.json", "r") as input:
            conditions = []
            for line in input:
                conditions.append(Condition(json.loads(line)))
            return conditions
    conditions = _conditions(dbgap_client)
    with open("dbgap_conditions.json", "w") as output:
        for condition in conditions:
            json.dump(condition.as_json(), output, separators=(',', ':'))
            output.write("\n")
    return conditions


def _validate_dbgap_conditions(dbgap_conditions):
    """There should be many Conditions."""
    assert dbgap_conditions, "Should return conditions"
    assert len(dbgap_conditions) > 0, "Should return conditions"


# def test_dbgap_conditions(dbgap_conditions):
#     """There should be many Conditions."""
#     _validate_dbgap_conditions(dbgap_conditions)


def _study_condition_systems(studies):
    systems = defaultdict(set)
    for research_study in studies:
        if research_study.resource_type is not "ResearchStudy":
            continue
        research_study_id = f"ResearchStudy/{research_study.id}"
        if not research_study.condition:
             systems[None].add(research_study_id)
             continue
        for condition in research_study.condition:
            if not condition.coding:
                systems[None].add(research_study_id)
                continue
            for coding in condition.coding:
                systems[coding.system].add(research_study_id)
    return systems


@pytest.fixture
def dbgap_study_condition_systems(dbgap_research_studies_with_observations):
    """There should be many a limited set of systems."""
    return _study_condition_systems(dbgap_research_studies_with_observations)


@pytest.fixture
def kidsfirst_study_condition_systems(kids_first_research_studies_with_observations):
    return _study_condition_systems(kids_first_research_studies_with_observations)


@pytest.fixture
def anvil_study_condition_systems(anvil_research_studies_with_observations):
    return _study_condition_systems(anvil_research_studies_with_observations)

# def test_condition_coding(kidsfirst_study_condition_systems, dbgap_study_condition_systems, anvil_study_condition_systems):
#     """We should use the same ontologies.
#     While all three systems use ResearchStudy.condition to indicate the condition that is the focus of the study. 
#     All three have a significant number of studies without a populated `condition`.
#     In addition, all three systems use different ontologies:
#     * kidsfirst ['http://snomed.info/sct', None]
#     * dbgap ['urn:oid:2.16.840.1.113883.6.177', 'https://dbgap-api.ncbi.nlm.nih.gov/fhir/x1/NamingSystem/MeshEntryTerm', 'https://uts.nlm.nih.gov/metathesaurus.html', None]
#     * anvil ['http://purl.obolibrary.org/obo/doid.owl', None]
#     """

#     from pprint import pprint
#     pprint({'kidsfirst': kidsfirst_study_condition_systems.keys()})
#     pprint({'dbgap': dbgap_study_condition_systems.keys()})
#     pprint({'anvil': anvil_study_condition_systems.keys()})

#     # only show the length of studies that match
#     for s in [kidsfirst_study_condition_systems, dbgap_study_condition_systems, anvil_study_condition_systems]:
#         for k in s:
#             s[k] = len(s[k])

#     pprint({'kidsfirst': kidsfirst_study_condition_systems})
#     pprint({'dbgap': dbgap_study_condition_systems})
#     pprint({'anvil': anvil_study_condition_systems})

#     kidsfirst = set(kidsfirst_study_condition_systems.keys())
#     dbgap = set(dbgap_study_condition_systems.keys())
#     anvil = set(anvil_study_condition_systems.keys())
#     assert kidsfirst.issubset(dbgap) and kidsfirst.issubset(anvil), f"{kidsfirst} is not a subset of {dbgap} {anvil}"

def test_anvil_condition(anvil_research_studies_with_observations):
    from pprint import pprint
    pprint(anvil_research_studies_with_observations)
    assert False


import os
import json
import pytest
import yaml
from click.testing import CliRunner
from anvil.etl.anvil_etl import cli
from importlib_metadata import distribution
import logging
from anvil.etl.utilities.shell_helper import PROPERTIES_LIST

logger = logging.getLogger(__name__)


def test_help():
    """Simply check options and commands."""
    runner = CliRunner()
    result = runner.invoke(cli, ['--help'])
    assert result.exit_code == 0
    # options
    for k in "--log-level,--output_path,--config_path".split(","):
        assert k in result.output
    # commands
    for k in "extract,transform,load,utility,version".split(","):
        assert k in result.output


def test_version():
    """Version should match."""
    runner = CliRunner()
    result = runner.invoke(cli, ['version'])
    assert result.exit_code == 0, result.output
    dist = distribution('pyAnVIL')
    assert dist.version in result.output


def test_config_format():
    """Should be able to print config in json or yaml."""
    runner = CliRunner()
    # json is default
    result = runner.invoke(cli, ['utility', 'config'])
    assert result.exit_code == 0
    assert json.loads(result.output), result.output

    # specify yaml
    result = runner.invoke(cli, ['utility', 'config', '--format', 'yaml'])
    assert result.exit_code == 0, result.output
    assert yaml.safe_load(result.output), result.output


def test_utility_env():
    """Should be able to print export variables."""
    runner = CliRunner()
    # json is default
    result = runner.invoke(cli, ['utility', 'env'])
    assert result.exit_code == 0, result.output
    for p in PROPERTIES_LIST:
        assert f"export {p}=" in result.output, result.output
    logger.debug("All expected properties present.")


@pytest.mark.skipif(reason="Setting environmental variables to arbitrary values interferes with downstream tests.")
def test_utility_existing_env():
    """Should respect existing variables."""
    runner = CliRunner()
    for p in PROPERTIES_LIST:
        os.environ[p] = "FOO"
    result = runner.invoke(cli, ['utility', 'env'])
    assert result.exit_code == 0, result.output
    for p in PROPERTIES_LIST:
        assert f"export {p}=FOO" in result.output, result.output
    logger.debug("Existing property values respected.")




"""Test terra wrapper."""

from anvil.util.reconciler import Reconciler
import logging

# logging.getLogger('anvil.test_terra').setLevel(logging.DEBUG)
logger = logging.getLogger('anvil.test_terra')


def test_reconciler_workspaces(user_project, namespaces, avro_path, terra_output_path, drs_output_path, project_pattern='.*CCDG.*'):
    """Ensure terra."""
    reconciler = Reconciler('CCDG', user_project, namespaces, project_pattern, avro_path, terra_output_path, drs_output_path)
    assert reconciler, "MUST create reconciler"
    assert len(reconciler.workspaces) == 338, "MUST have at least expected number of workspaces"
    for w in reconciler.workspaces:
        assert 'ccdg' in w.attributes['workspace']['name'].lower(), "name should include ccdg"


def test_reconciler_blobs(user_project, namespaces, avro_path, terra_output_path, drs_output_path, project_pattern='.*CCDG.*'):
    """Ensure terra."""
    reconciler = Reconciler('CCDG', user_project, namespaces, project_pattern, avro_path, terra_output_path, drs_output_path)
    attributes = reconciler.workspaces[0].attributes
    dashboard_view = reconciler.workspaces[0].dashboard_view
    assert sorted(dashboard_view.keys()) == ['createdDate', 'data_category', 'data_type', 'file_histogram', 'files', 'lastModified', 'nodes', 'problems', 'project_id', 'public', 'size']

    assert dashboard_view.size, f"MUST return blobs {attributes['workspace']['name']} {dashboard_view}"



import os
from anvil.clients.fhir_client import FHIRClient
from anvil.clients.smart_auth import GoogleFHIRAuth
import fhirclient.models.researchstudy as rs


def test_query():
    settings = {
        'app_id': 'my_web_app',
        'api_base': 'https://healthcare.googleapis.com/v1beta1/projects/fhir-test-16-342800/locations/us-west2/datasets/anvil-test/fhirStores/public/fhir'
    }
    smart = FHIRClient(settings=settings, auth=GoogleFHIRAuth())
    smart.prepare()
    assert smart.ready, "server should be ready"
    # search for all ResearchStudy
    ids = [s.id for s in rs.ResearchStudy.where(struct={}).perform_resources(smart.server)]
    print(ids)
    assert len(ids) > 0, "Should have at least one ResearchStudy"



from anvil.clients.fhir_client import FHIRClient, DispatchingFHIRClient
from anvil.clients.smart_auth import GoogleFHIRAuth
import fhirclient.models.researchstudy as rs


def test_query_simple():
    settings = {
        'app_id': 'my_web_app',
        'api_base': 'https://healthcare.googleapis.com/v1beta1/projects/fhir-test-16-342800/locations/us-west2/datasets/anvil-test/fhirStores/public/fhir'
    }
    smart = FHIRClient(settings=settings, auth=GoogleFHIRAuth())
    smart.prepare()
    assert smart.ready, "server should be ready"
    # search for all ResearchStudy
    ids = [s.id for s in rs.ResearchStudy.where(struct={}).perform_resources(smart.server)]
    print(ids)
    assert len(ids) > 0, "Should have at least one ResearchStudy"


def test_query_request_jsonl():
    settings = {
        'app_id': 'my_web_app',
        'api_base': 'https://healthcare.googleapis.com/v1beta1/projects/fhir-test-16-342800/locations/us-west2/datasets/anvil-test/fhirStores/public/fhir'
    }
    client = FHIRClient(settings=settings, auth=GoogleFHIRAuth())
    client.prepare()
    assert client.ready, "server should be ready"
    response = client.server.request_json(path=f'{client.server.base_uri}ResearchStudy?_elements=id&_count=1000')
    assert 'entry' in response, response


def test_dispatching_query_get():
    settings = {
        'app_id': 'my_web_app',
        'api_bases': [
            'https://healthcare.googleapis.com/v1beta1/projects/fhir-test-16-342800/locations/us-west2/datasets/anvil-test/fhirStores/public/fhir',
            'https://healthcare.googleapis.com/v1beta1/projects/fhir-test-16-342800/locations/us-west2/datasets/anvil-test/fhirStores/pending/fhir',
        ]
    }
    client = DispatchingFHIRClient(settings=settings, auth=GoogleFHIRAuth())
    client.prepare()
    assert client.ready, "server should be ready"
    assert 'fhir-test-16-342800' in client.server.base_uri
    response = client.server._get(path='ResearchStudy?_elements=id&_count=1000')
    assert response.json()


def test_dispatching_worker():
    settings = {
        'app_id': 'my_web_app',
        'api_bases': [
            'https://healthcare.googleapis.com/v1beta1/projects/fhir-test-16-342800/locations/us-west2/datasets/anvil-test/fhirStores/public/fhir',
            'https://healthcare.googleapis.com/v1beta1/projects/fhir-test-16-342800/locations/us-west2/datasets/anvil-test/fhirStores/pending/fhir',
        ]
    }
    client = DispatchingFHIRClient(settings=settings, auth=GoogleFHIRAuth())
    client.prepare()
    assert client.ready, "server should be ready"
    assert 'fhir-test-16-342800' in client.server.base_uri

    def worker(server):
        """Callback handler"""
        response = server._get(path='ResearchStudy?_elements=id&_count=1000')
        assert response.json()
        return response.json()

    results = client.dispatch(worker)
    for r in results:
        assert 'entry' in r


